import re
from typing import Callable, Iterable, Optional, Sequence
from .. import base
from ..spacing import Newline, Whitespace

_SPACING_GROUP_RE = re.compile(r'([ \t]+)|(\r*\n)')


def _tokens_to_text(tokens: Iterable[base.RawTokenModel]) -> str:
    return ''.join(token.raw_text for token in tokens)


def _text_to_tokens(text: str) -> Iterable[base.RawTokenModel]:
    for whitespace, newline in _SPACING_GROUP_RE.findall(text):
        if whitespace:
            yield Whitespace.from_raw_text(whitespace)
        if newline:
            yield Newline.from_raw_text(newline)


def _find_spacing(
        token: Optional[base.RawTokenModel],
        succ: Callable[[base.RawTokenModel], Optional[base.RawTokenModel]],
) -> list[base.RawTokenModel]:
    tokens: list[base.RawTokenModel] = []
    while token is not None and not token.raw_text:
        token = succ(token)
    # must not interleave with special tokens to avoid removing them in spacing update.
    while isinstance(token, Newline | Whitespace):
        if token.raw_text:
            tokens.append(token)
        token = succ(token)
    return tokens


class SpacingAccessorsMixin(base.RawModel):

    @property
    def raw_spacing_before(self) -> tuple[base.RawTokenModel, ...]:
        if self.token_store is None:
            return ()
        return tuple(reversed(_find_spacing(
                self.token_store.get_prev(self.first_token),
                self.token_store.get_prev)))

    @raw_spacing_before.setter
    def raw_spacing_before(self, tokens: tuple[base.RawTokenModel, ...]) -> None:
        if self.token_store is None:
            raise ValueError('Cannot set spacing without a token store.')
        current_tokens = self.raw_spacing_before
        if current_tokens:
            self.token_store.splice(tokens, current_tokens[0], current_tokens[-1])
        else:
            self.token_store.insert_before(self.first_token, tokens)

    @property
    def spacing_before(self) -> str:
        return _tokens_to_text(self.raw_spacing_before)

    @spacing_before.setter
    def spacing_before(self, value: str) -> None:
        self.raw_spacing_before = tuple(_text_to_tokens(value))

    @property
    def raw_spacing_after(self) -> Sequence[base.RawTokenModel]:
        if self.token_store is None:
            return ()
        return tuple(_find_spacing(
                self.token_store.get_next(self.last_token),
                self.token_store.get_next))

    @raw_spacing_after.setter
    def raw_spacing_after(self, tokens: Sequence[base.RawTokenModel]) -> None:
        if self.token_store is None:
            raise ValueError('Cannot set spacing without a token store.')
        current_tokens = self.raw_spacing_after
        if current_tokens:
            self.token_store.splice(tokens, current_tokens[0], current_tokens[-1])
        else:
            self.token_store.insert_after(self.last_token, tokens)

    @property
    def spacing_after(self) -> str:
        return _tokens_to_text(self.raw_spacing_after)

    @spacing_after.setter
    def spacing_after(self, value: str) -> None:
        self.raw_spacing_after = tuple(_text_to_tokens(value))



import bisect
import dataclasses
from typing import Any, Collection, Generic, Iterable, Iterator, Optional, Type, TypeVar

_T = TypeVar('_T', bound='Token')
_Self = TypeVar('_Self', bound='TokenStore')


@dataclasses.dataclass(frozen=True)
class Position:
    position: int
    line: int
    column: int

    def __add__(self, other: Any) -> 'Position':
        if not isinstance(other, Position):
            raise TypeError(
                f"unsupported operand type(s) for +: {type(self).__name__!r} "
                f"and {type(other).__name__!r}.")
        if other.line:
            column = other.column
        else:
            column = self.column + other.column
        return Position(
            position=self.position + other.position,
            line=self.line + other.line,
            column=column)  # not commutative!


class Token:
    _raw_text: str
    store_handle: Optional['_Handle']

    def __init__(self, raw_text: str) -> None:
        self._raw_text = raw_text
        self.store_handle = None

    @property
    def raw_text(self) -> str:
        return self._raw_text

    @raw_text.setter
    def raw_text(self, value: str) -> None:
        self._update_raw_text(value)

    # property setter override is painful. use setter method instead.
    def _update_raw_text(self, value: str) -> None:
        self._raw_text = value
        if self.store_handle:
            self.store_handle.store.update(self)


@dataclasses.dataclass(eq=False)
class _Handle:
    store: 'TokenStore'
    index: int
    position: Position


def _check_store_handle(token: Token) -> _Handle:
    if not token.store_handle:
        raise ValueError('Token is not in a store.')
    return token.store_handle


def _token_size(token: Token) -> Position:
    return Position(
        position=len(token.raw_text),
        line=token.raw_text.count('\n'),
        column=len(token.raw_text) - token.raw_text.rfind('\n') - 1)


class TokenStore(Generic[_T]):
    """Storage for tokens allowing insertion, deletion and lookup by position."""

    _tokens: list[_T]
    _end: Position

    def __init__(self) -> None:
        self._tokens = []
        self._end = Position(0, 0, 0)

    @classmethod
    def from_tokens(cls: Type[_Self], tokens: Iterable[_T]) -> _Self:
        for token in tokens:
            if token.store_handle:
                raise ValueError('Token already in a store.')
        store = cls()
        for token in tokens:
            token.store_handle = _Handle(
                store=store, index=0, position=Position(0, 0, 0))
        store._tokens.extend(tokens)
        store._update_token_handles()
        return store

    def _update_token_handles(self, from_index: int = 0) -> None:
        if from_index:
            prev = self._tokens[from_index - 1]
            position = _check_store_handle(prev).position + _token_size(prev)
        else:
            position = Position(0, 0, 0)
        for index in range(from_index, len(self._tokens)):
            token = self._tokens[index]
            handle = _check_store_handle(token)
            handle.position = position
            handle.index = index
            position += _token_size(token)
        self._end = position

    def _splice(self, tokens: Collection[_T], start: int, end: int) -> None:
        for token in tokens:
            handle = token.store_handle
            if handle and not (handle.store is self and start <= handle.index <= end):
                raise ValueError('Token already in a store.')
        for token in self._tokens[start:end]:
            token.store_handle = None
        for token in tokens:
            token.store_handle = _Handle(store=self, index=0, position=Position(0, 0, 0))
        self._tokens[start:end] = tokens
        self._update_token_handles(start)

    def splice(self, tokens: Collection[_T], ref: Optional[_T], del_end: Optional[_T] = None) -> None:
        start = _check_store_handle(ref).index if ref else 0
        end = _check_store_handle(del_end).index + 1 if del_end else start
        self._splice(tokens, start, end)

    def insert_after(self, ref: Optional[_T], tokens: Collection[_T]) -> None:
        if ref is None:
            start = 0
        else:
            start = _check_store_handle(ref).index + 1
        self._splice(tokens, start, start)

    def insert_before(self, ref: Optional[_T], tokens: Collection[_T]) -> None:
        self.splice(tokens, ref)

    def update(self, token: _T) -> None:
        handle = _check_store_handle(token)
        self._update_token_handles(handle.index + 1)

    def replace(self, token: _T, repl: _T) -> None:
        self.splice([repl], token, token)

    def remove(self, start: _T, end: Optional[_T] = None) -> None:
        self.splice([], start, end or start)

    def iter(self, start: _T, end: _T) -> Iterator[_T]:
        start_idx = _check_store_handle(start).index
        end_idx = _check_store_handle(end).index
        yield from self._tokens[start_idx:end_idx + 1]

    def get_index(self, token: _T) -> int:
        handle = _check_store_handle(token)
        return handle.index

    def get_position(self, token: _T) -> Position:
        handle = _check_store_handle(token)
        return handle.position

    def get_by_position(self, position: int) -> _T:
        index = bisect.bisect_left(
            self._tokens, position, key=lambda t: _check_store_handle(t).position.position)
        token = next((token for token in self._tokens[index:] if token.raw_text), None)
        if token is None or _check_store_handle(token).position.position != position:
            raise KeyError(f'No token found at position {position}.')
        return token

    def find_by_line(self, line: int) -> Iterator[_T]:  # zero-based
        index = bisect.bisect_left(
            self._tokens, line, key=lambda t: _check_store_handle(t).position.line)
        while index < len(self._tokens):
            token = self._tokens[index]
            if not token.raw_text:
                continue
            if _check_store_handle(token).position.line >= line:
                break
            yield self._tokens[index]

    def get_prev(self, token: _T) -> Optional[_T]:
        index = _check_store_handle(token).index
        return self._tokens[index - 1] if index else None

    def get_next(self, token: _T) -> Optional[_T]:
        index = _check_store_handle(token).index
        return self._tokens[index + 1] if index + 1 < len(self._tokens) else None

    def get_first(self) -> Optional[_T]:
        return next(iter(self._tokens), None)

    def get_last(self) -> Optional[_T]:
        return next(reversed(self._tokens), None)

    def __iter__(self) -> Iterator[_T]:
        yield from self._tokens

    def __len__(self) -> int:
        return len(self._tokens)

    @ property
    def size(self) -> int:
        return self._end.position



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import decimal
from typing import Optional, Type, TypeVar, final
from .. import base, internal
from ..currency import Currency
from ..number_expr import NumberExpr
from ..spacing import Whitespace

_Self = TypeVar('_Self', bound='TotalPrice')


@internal.token_model
class AtAt(internal.SimpleDefaultRawTokenModel):
    RULE = 'ATAT'
    DEFAULT = '@@'


@internal.tree_model
class TotalPrice(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'total_price'

    _label = internal.required_field[AtAt]()
    _number = internal.optional_left_field[NumberExpr](separators=(Whitespace.from_default(),))
    _currency = internal.optional_left_field[Currency](separators=(Whitespace.from_default(),))

    @internal.custom_property
    def _number_pivot(self) -> base.RawTokenModel:
        return self._label.last_token

    @internal.custom_property
    def _currency_pivot(self) -> base.RawTokenModel:
        return (self._number and self._number.last_token) or self._label.last_token

    raw_number = internal.optional_node_property(_number, _number_pivot)
    raw_currency = internal.optional_node_property(_currency, _currency_pivot)

    number = internal.optional_decimal_property(raw_number, NumberExpr)
    currency = internal.optional_string_property(raw_currency, Currency)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            label: AtAt,
            number: Optional[NumberExpr],
            currency: Optional[Currency],
    ):
        super().__init__(token_store)
        self._label = label
        self._number = number
        self._currency = currency

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._label.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return (self._currency and self._currency.last_token) or (self._number and self._number.last_token) or self._label.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._label.clone(self._label, token_store, token_transformer),
            type(self)._number.clone(self._number, token_store, token_transformer),
            type(self)._currency.clone(self._currency, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._label = type(self)._label.reattach(self._label, token_store, token_transformer)
        self._number = type(self)._number.reattach(self._number, token_store, token_transformer)
        self._currency = type(self)._currency.reattach(self._currency, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, TotalPrice)
            and self._label == other._label
            and self._number == other._number
            and self._currency == other._currency
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            number: Optional[NumberExpr],
            currency: Optional[Currency],
    ) -> _Self:
        label = AtAt.from_default()
        tokens = [
            *label.detach(),
            *cls._number.detach_with_separators(number),
            *cls._currency.detach_with_separators(currency),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._label.reattach(label, token_store)
        cls._number.reattach(number, token_store)
        cls._currency.reattach(currency, token_store)
        return cls(token_store, label, number, currency)

    @classmethod
    def from_value(
            cls: Type[_Self],
            number: Optional[decimal.Decimal],
            currency: Optional[str],
    ) -> _Self:
        return cls.from_children(
            number=NumberExpr.from_value(number) if number is not None else None,
            currency=Currency.from_value(currency) if currency is not None else None,
        )

    def auto_claim_comments(self) -> None:
        type(self)._currency.auto_claim_comments(self._currency)
        type(self)._number.auto_claim_comments(self._number)



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import Iterable, Type, TypeVar, final
from .. import base, internal
from ..cost_component import CostComponent
from ..punctuation import Comma
from ..spacing import Whitespace

_Self = TypeVar('_Self', bound='UnitCost')


@internal.token_model
class LeftBrace(internal.SimpleDefaultRawTokenModel):
    RULE = 'LEFT_BRACE'
    DEFAULT = '{'


@internal.token_model
class RightBrace(internal.SimpleDefaultRawTokenModel):
    RULE = 'RIGHT_BRACE'
    DEFAULT = '}'


@internal.tree_model
class UnitCost(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'unit_cost'

    _left_brace = internal.required_field[LeftBrace]()
    _components = internal.repeated_field[CostComponent](separators=(Comma.from_default(), Whitespace.from_default()), separators_before=())
    _right_brace = internal.required_field[RightBrace]()

    raw_components = internal.repeated_node_property(_components)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            left_brace: LeftBrace,
            repeated_components: internal.Repeated[CostComponent],
            right_brace: RightBrace,
    ):
        super().__init__(token_store)
        self._left_brace = left_brace
        self._components = repeated_components
        self._right_brace = right_brace

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._left_brace.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._right_brace.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._left_brace.clone(self._left_brace, token_store, token_transformer),
            type(self)._components.clone(self._components, token_store, token_transformer),
            type(self)._right_brace.clone(self._right_brace, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._left_brace = type(self)._left_brace.reattach(self._left_brace, token_store, token_transformer)
        self._components = type(self)._components.reattach(self._components, token_store, token_transformer)
        self._right_brace = type(self)._right_brace.reattach(self._right_brace, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, UnitCost)
            and self._left_brace == other._left_brace
            and self._components == other._components
            and self._right_brace == other._right_brace
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            components: Iterable[CostComponent],
    ) -> _Self:
        left_brace = LeftBrace.from_default()
        repeated_components = cls._components.create_repeated(components)
        right_brace = RightBrace.from_default()
        tokens = [
            *left_brace.detach(),
            *cls._components.detach_with_separators(repeated_components),
            *right_brace.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._left_brace.reattach(left_brace, token_store)
        cls._components.reattach(repeated_components, token_store)
        cls._right_brace.reattach(right_brace, token_store)
        return cls(token_store, left_brace, repeated_components, right_brace)

    def auto_claim_comments(self) -> None:
        self.raw_components.auto_claim_comments()



import re
from . import internal


@internal.token_model
class EscapedString(internal.SingleValueRawTokenModel[str]):
    RULE = 'ESCAPED_STRING'
    # See: https://github.com/beancount/beancount/blob/d841487ccdda04c159de86b1186e7c2ea997a3e2/beancount/parser/tokens.c#L102
    __ESCAPE_MAP = {
        '\n': 'n',
        '\t': 't',
        '\r': 'r',
        '\f': 'f',
        '\b': 'b',
        '"': '"',
        '\\': '\\',
    }
    __ESCAPE_PATTERN = re.compile(r'[\\"]')
    __ESCAPE_PATTERN_AGGRESSIVE = re.compile(
        '|'.join(map(re.escape, __ESCAPE_MAP.keys())))
    __UNESCAPE_MAP = {value: key for key, value in __ESCAPE_MAP.items()}
    __UNESCAPE_PATTERN = re.compile(r'\\(.)')

    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        return cls.unescape(raw_text[1:-1])
    
    @classmethod
    def _format_value(cls, value: str) -> str:
        return f'"{cls.escape(value)}"'

    @classmethod
    def escape(cls, s: str, aggressive: bool = False) -> str:
        pattern = cls.__ESCAPE_PATTERN_AGGRESSIVE if aggressive else cls.__ESCAPE_PATTERN
        return re.sub(
            pattern,
            lambda c: '\\' + cls.__ESCAPE_MAP[c.group(0)],
            s)

    @classmethod
    def unescape(cls, s: str) -> str:
        return re.sub(
            cls.__UNESCAPE_PATTERN,
            lambda c: cls.__UNESCAPE_MAP.get(c.group(1), c.group(1)),
            s)



import decimal
from . import internal


@internal.token_model
class Number(internal.SingleValueRawTokenModel[decimal.Decimal]):
    RULE = 'NUMBER'

    @classmethod
    def _parse_value(cls, raw_text: str) -> decimal.Decimal:
        return decimal.Decimal(raw_text.replace(',', ''))
    
    @classmethod
    def _format_value(cls, value: decimal.Decimal) -> str:
        return str(value)



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import Optional, Type, TypeVar, final
from .. import base, internal
from ..block_comment import BlockComment
from ..ignored import Ignored
from ..punctuation import Eol
from ..spacing import Newline

_Self = TypeVar('_Self', bound='IgnoredLine')


@internal.tree_model
class IgnoredLine(internal.SurroundingCommentsMixin, base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'ignored_line'

    _ignored = internal.required_field[Ignored]()
    _eol = internal.required_field[Eol]()

    @internal.custom_property
    def _leading_comment_pivot(self) -> base.RawTokenModel:
        return self._ignored.first_token

    @internal.custom_property
    def _trailing_comment_pivot(self) -> base.RawTokenModel:
        return self._eol.last_token

    raw_leading_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._leading_comment, _leading_comment_pivot)
    raw_ignored = internal.required_node_property(_ignored)
    raw_trailing_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._trailing_comment, _trailing_comment_pivot)

    leading_comment = internal.optional_string_property(raw_leading_comment, BlockComment)
    trailing_comment = internal.optional_string_property(raw_trailing_comment, BlockComment)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            leading_comment: Optional[BlockComment],
            ignored: Ignored,
            eol: Eol,
            trailing_comment: Optional[BlockComment],
    ):
        super().__init__(token_store)
        self._leading_comment = leading_comment
        self._ignored = ignored
        self._eol = eol
        self._trailing_comment = trailing_comment

    @property
    def first_token(self) -> base.RawTokenModel:
        return (self._leading_comment and self._leading_comment.first_token) or self._ignored.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return (self._trailing_comment and self._trailing_comment.last_token) or self._eol.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._leading_comment.clone(self._leading_comment, token_store, token_transformer),
            type(self)._ignored.clone(self._ignored, token_store, token_transformer),
            type(self)._eol.clone(self._eol, token_store, token_transformer),
            type(self)._trailing_comment.clone(self._trailing_comment, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._leading_comment = type(self)._leading_comment.reattach(self._leading_comment, token_store, token_transformer)
        self._ignored = type(self)._ignored.reattach(self._ignored, token_store, token_transformer)
        self._eol = type(self)._eol.reattach(self._eol, token_store, token_transformer)
        self._trailing_comment = type(self)._trailing_comment.reattach(self._trailing_comment, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, IgnoredLine)
            and self._leading_comment == other._leading_comment
            and self._ignored == other._ignored
            and self._eol == other._eol
            and self._trailing_comment == other._trailing_comment
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            ignored: Ignored,
            *,
            leading_comment: Optional[BlockComment] = None,
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        eol = Eol.from_default()
        tokens = [
            *cls._leading_comment.detach_with_separators(leading_comment),
            *ignored.detach(),
            *eol.detach(),
            *cls._trailing_comment.detach_with_separators(trailing_comment),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._leading_comment.reattach(leading_comment, token_store)
        cls._ignored.reattach(ignored, token_store)
        cls._eol.reattach(eol, token_store)
        cls._trailing_comment.reattach(trailing_comment, token_store)
        return cls(token_store, leading_comment, ignored, eol, trailing_comment)

    def auto_claim_comments(self) -> None:
        self.claim_leading_comment(ignore_if_already_claimed=True)
        self.claim_trailing_comment(ignore_if_already_claimed=True)
        type(self)._trailing_comment.auto_claim_comments(self._trailing_comment)
        type(self)._ignored.auto_claim_comments(self._ignored)
        type(self)._leading_comment.auto_claim_comments(self._leading_comment)



import datetime
import decimal
from typing import Generic, Optional, Type, TypeVar, overload
from . import internal
from .base import RawTreeModel
from .bool import Bool
from .date import Date
from .escaped_string import EscapedString
from .number_expr import NumberExpr
from .meta_value import MetaRawValue, MetaValue

_U = TypeVar('_U', bound=RawTreeModel)


# Not inside internal.value_properties to avoid circular dependencies.
class optional_meta_value_property(Generic[_U]):
    def __init__(self, inner_property: internal.base_rw_property[Optional[MetaRawValue], _U]):
        self.inner_property = inner_property

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> Optional[MetaValue]:
        raw_value = self.inner_property.__get__(instance, owner)
        if isinstance(raw_value, EscapedString | Date | NumberExpr | Bool):
            return raw_value.value
        return raw_value

    def __set__(self, instance: _U, value: Optional[MetaValue | MetaRawValue]) -> None:
        current_raw = self.inner_property.__get__(instance, None)
        if not update_value(current_raw, value):
            self.inner_property.__set__(instance, from_value(value))


def update_value(raw_value: Optional[MetaRawValue], value: Optional[MetaValue | MetaRawValue]) -> bool:
    match raw_value, value:
        case EscapedString() as m, str() as v:
            m.value = v
        case Date() as m, datetime.date() as v:
            m.value = v
        case NumberExpr() as m, decimal.Decimal() as v:
            m.value = v
        case Bool() as m, bool() as v:
            m.value = v
        case _:
            return False
    return True


def from_value(value: Optional[MetaValue | MetaRawValue]) -> Optional[MetaRawValue]:
    match value:
        case str():
            return EscapedString.from_value(value)
        case datetime.date():
            return Date.from_value(value)
        case decimal.Decimal():
            return NumberExpr.from_value(value)
        case bool():
            return Bool.from_value(value)
    return value



from . import internal


@internal.token_model
class TransactionFlag(internal.SingleValueRawTokenModel[str]):
    RULE = 'TRANSACTION_FLAG'

    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        if raw_text == 'txn':
            return '*'
        return raw_text

    @classmethod
    def _format_value(cls, value: str) -> str:
        return value



from . import base
from . import internal
from .generated import file
from .generated.file import Directive


@internal.tree_model
class File(file.File):

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._token_store.get_first() or self._directives.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._token_store.get_last() or self._directives.last_token



from ..base import RawTokenModel


class Placeholder(RawTokenModel):
    RULE = 'PLACEHOLDER'

    @classmethod
    def from_default(cls) -> 'Placeholder':
        return cls('')

    def _clone(self) -> 'Placeholder':
        return Placeholder('')



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import decimal
from typing import Type, TypeVar, final
from .. import base, internal
from ..currency import Currency
from ..number_expr import NumberExpr
from ..spacing import Whitespace

_Self = TypeVar('_Self', bound='Amount')


@internal.tree_model
class Amount(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'amount'

    _number = internal.required_field[NumberExpr]()
    _currency = internal.required_field[Currency]()

    raw_number = internal.required_node_property(_number)
    raw_currency = internal.required_node_property(_currency)

    number = internal.required_value_property(raw_number)
    currency = internal.required_value_property(raw_currency)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            number: NumberExpr,
            currency: Currency,
    ):
        super().__init__(token_store)
        self._number = number
        self._currency = currency

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._number.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._currency.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._number.clone(self._number, token_store, token_transformer),
            type(self)._currency.clone(self._currency, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._number = type(self)._number.reattach(self._number, token_store, token_transformer)
        self._currency = type(self)._currency.reattach(self._currency, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, Amount)
            and self._number == other._number
            and self._currency == other._currency
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            number: NumberExpr,
            currency: Currency,
    ) -> _Self:
        tokens = [
            *number.detach(),
            Whitespace.from_default(),
            *currency.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._number.reattach(number, token_store)
        cls._currency.reattach(currency, token_store)
        return cls(token_store, number, currency)

    @classmethod
    def from_value(
            cls: Type[_Self],
            number: decimal.Decimal,
            currency: str,
    ) -> _Self:
        return cls.from_children(
            number=NumberExpr.from_value(number),
            currency=Currency.from_value(currency),
        )

    def auto_claim_comments(self) -> None:
        type(self)._currency.auto_claim_comments(self._currency)
        type(self)._number.auto_claim_comments(self._number)



import copy
from typing import Generic, Iterable, Optional, Type, TypeVar
from .. import base
from .placeholder import Placeholder

_M = TypeVar('_M', bound=base.RawModel)
_Self = TypeVar('_Self', bound='Repeated')

class Repeated(base.RawTreeModel, Generic[_M]):
    def __init__(
            self,
            token_store: base.TokenStore,
            items: Iterable[_M],
            placeholder: Placeholder,
            inferred_indent: Optional[str] = None,
    ) -> None:
        super().__init__(token_store)
        self.items = list(items)
        self._placeholder = placeholder
        self._inferred_indent = inferred_indent

    @property
    def placeholder(self) -> Placeholder:
        return self._placeholder

    @property
    def inferred_indent(self) -> Optional[str]:
        return self._inferred_indent

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._placeholder

    @property
    def last_token(self) -> base.RawTokenModel:
        if self.items:
            return self.items[-1].last_token
        return self._placeholder

    def _eq(self, other: base.RawTreeModel) -> bool:
        return isinstance(other, Repeated) and self.items == other.items

    @classmethod
    def from_children(
            cls: Type[_Self],
            items: Iterable[_M],
            *,
            separators: tuple[base.RawTokenModel, ...],
            separators_before: Optional[tuple[base.RawTokenModel, ...]] = None,
            indent: Optional[str] = None,
    ) -> _Self:
        placeholder = Placeholder.from_default()
        items = list(items)
        tokens: list[base.RawTokenModel] = [placeholder]
        for i, item in enumerate(items):
            if i == 0 and separators_before is not None:
                tokens.extend(copy.deepcopy(separators_before))
            else:
                tokens.extend(copy.deepcopy(separators))
            tokens.extend(item.detach())
        token_store = base.TokenStore.from_tokens(tokens)
        for item in items:
            item.reattach(token_store)
        return cls(token_store, items, placeholder, indent)

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            (item.clone(token_store, token_transformer) for item in self.items),
            self.placeholder.clone(token_store, token_transformer),
            self._inferred_indent)

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self.items = [item.reattach(token_store, token_transformer) for item in self.items]
        self._placeholder = self._placeholder.reattach(token_store, token_transformer)

    def auto_claim_comments(self) -> None:
        for item in reversed(self.items):
            item.auto_claim_comments()



import datetime
import itertools
from typing import Iterable, Mapping, Optional, Type, TypeVar
from . import internal, meta_item_internal
from .date import Date
from .account import Account
from .escaped_string import EscapedString
from .inline_comment import InlineComment
from .link import Link
from .tag import Tag
from .generated import document
from .generated.document import DocumentLabel
from .meta_value import MetaRawValue, MetaValue

_Self = TypeVar('_Self', bound='Document')


@internal.tree_model
class Document(document.Document):
    tags = internal.repeated_string_property(document.Document.raw_tags_links, Tag)
    links = internal.repeated_string_property(document.Document.raw_tags_links, Link)

    @classmethod
    def from_value(
            cls: Type[_Self],
            date: datetime.date,
            account: str,
            filename: str,
            *,
            tags: Iterable[str] = (),
            links: Iterable[str] = (),
            inline_comment: Optional[str] = None,
            meta: Optional[Mapping[str, MetaValue | MetaRawValue]] = None,
    ) -> _Self:
        return cls.from_children(
            date=Date.from_value(date),
            account=Account.from_value(account),
            filename=EscapedString.from_value(filename),
            tags_links=itertools.chain(map(Tag.from_value, tags), map(Link.from_value, links)),
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            meta=meta_item_internal.from_mapping(meta) if meta is not None else (),
        )



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import TYPE_CHECKING, Type, TypeVar, final
from .. import base, internal
from ..spacing import Whitespace
if TYPE_CHECKING:
  from ..number_add_expr import NumberAddExpr

_Self = TypeVar('_Self', bound='NumberParenExpr')


@internal.token_model
class LeftParen(internal.SimpleDefaultRawTokenModel):
    RULE = 'LEFT_PAREN'
    DEFAULT = '('


@internal.token_model
class RightParen(internal.SimpleDefaultRawTokenModel):
    RULE = 'RIGHT_PAREN'
    DEFAULT = ')'


@internal.tree_model
class NumberParenExpr(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'number_paren_expr'

    _left_paren = internal.required_field[LeftParen]()
    _inner_expr = internal.required_field['NumberAddExpr']()
    _right_paren = internal.required_field[RightParen]()

    raw_inner_expr = internal.required_node_property(_inner_expr)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            left_paren: LeftParen,
            inner_expr: 'NumberAddExpr',
            right_paren: RightParen,
    ):
        super().__init__(token_store)
        self._left_paren = left_paren
        self._inner_expr = inner_expr
        self._right_paren = right_paren

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._left_paren.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._right_paren.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._left_paren.clone(self._left_paren, token_store, token_transformer),
            type(self)._inner_expr.clone(self._inner_expr, token_store, token_transformer),
            type(self)._right_paren.clone(self._right_paren, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._left_paren = type(self)._left_paren.reattach(self._left_paren, token_store, token_transformer)
        self._inner_expr = type(self)._inner_expr.reattach(self._inner_expr, token_store, token_transformer)
        self._right_paren = type(self)._right_paren.reattach(self._right_paren, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, NumberParenExpr)
            and self._left_paren == other._left_paren
            and self._inner_expr == other._inner_expr
            and self._right_paren == other._right_paren
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            inner_expr: 'NumberAddExpr',
    ) -> _Self:
        left_paren = LeftParen.from_default()
        right_paren = RightParen.from_default()
        tokens = [
            *left_paren.detach(),
            Whitespace.from_default(),
            *inner_expr.detach(),
            Whitespace.from_default(),
            *right_paren.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._left_paren.reattach(left_paren, token_store)
        cls._inner_expr.reattach(inner_expr, token_store)
        cls._right_paren.reattach(right_paren, token_store)
        return cls(token_store, left_paren, inner_expr, right_paren)

    def auto_claim_comments(self) -> None:
        type(self)._inner_expr.auto_claim_comments(self._inner_expr)



from typing import Generic, ItemsView, Iterable, Iterator, KeysView, Mapping, MutableMapping, Optional, Type, TypeVar, ValuesView, no_type_check, overload
from .meta_item import MetaItem
from .meta_value import MetaRawValue, MetaValue
from . import base, internal

_V = TypeVar('_V')


class _Empty:
    pass


_EMPTY = _Empty()


class _DictView:
    def __init__(self, wrapper: internal.RepeatedValueWrapper[MetaItem, MetaItem]):
        self._wrapper = wrapper

    def __len__(self) -> int:
        return len(self._wrapper)


class RepeatedRawMetaKeysView(_DictView, KeysView[str]):
    def __iter__(self) -> Iterator[str]:
        for item in self._wrapper:
            yield item.key

    def __reversed__(self) -> Iterator[str]:
        for item in reversed(self._wrapper):
            yield item.key


class RepeatedRawMetaValuesView(_DictView, ValuesView[MetaItem]):
    def __iter__(self) -> Iterator[MetaItem]:
        return iter(self._wrapper)

    def __reversed__(self) -> Iterator[MetaItem]:
        return reversed(self._wrapper)


class RepeatedRawMetaItemsView(_DictView, ItemsView[str, MetaItem]):
    def __iter__(self) -> Iterator[tuple[str, MetaItem]]:
        for item in self._wrapper:
            yield item.key, item

    def __reversed__(self) -> Iterator[tuple[str, MetaItem]]:
        for item in reversed(self._wrapper):
            yield item.key, item


class RepeatedRawMetaItemWrapper(
        internal.RepeatedFilteredNodeWrapper[MetaItem],
        MutableMapping[str, MetaItem]):

    def __init__(
            self,
            raw_wrapper: internal.RepeatedNodeWithInterleavingCommentsWrapper[MetaItem],
    ) -> None:
        super().__init__(raw_wrapper, MetaItem)

    @overload
    def __getitem__(self, index: int) -> MetaItem:
        ...
    @overload
    def __getitem__(self, index: slice) -> list[MetaItem]:
        ...
    @overload
    def __getitem__(self, index: str) -> MetaItem:
        ...
    def __getitem__(self, index: int | slice | str) -> MetaItem | list[MetaItem] | MetaItem:
        if not isinstance(index, str):
            return super().__getitem__(index)
        for item in self:
            if item.key == index:
                return item
        raise KeyError(index)

    def __delitem__(self, index: int | slice | str) -> None:
        if not isinstance(index, str):
            return super().__delitem__(index)
        for i, item in enumerate(self):
            if item.key == index:
                return super().__delitem__(i)
        raise KeyError(index)

    @overload
    def __setitem__(self, index: int, value: MetaItem) -> None:
        ...
    @overload
    def __setitem__(self, index: slice, value: Iterable[MetaItem]) -> None:
        ...
    @overload
    def __setitem__(self, index: str, value: MetaItem) -> None:
        ...
    @no_type_check  # mypy does not seem to work well with overload
    def __setitem__(self, index: int | slice | str, value: MetaItem | Iterable[MetaItem]) -> None:
        if not isinstance(index, str):
            return super().__setitem__(index, value)
        for i, item in enumerate(self):
            if item.key == index:
                return super().__setitem__(i, value)
        self.append(value)

    @overload
    def pop(self, index: int = -1) -> MetaItem:
        ...
    @overload
    def pop(self, index: str) -> MetaItem:
        ...
    @overload
    def pop(self, index: str, default: _V) -> MetaItem | _V:
        ...
    def pop(self, index: int | str = -1, default: _V | _Empty = _EMPTY) -> MetaItem | _V:
        if not isinstance(index, str):
            return super().pop(index)
        for i, item in enumerate(self):
            if item.key == index:
                return super().pop(i)
        if not isinstance(default, _Empty):
            return default
        raise KeyError(index)

    def keys(self) -> KeysView[str]:
        return RepeatedRawMetaKeysView(self)

    def values(self) -> ValuesView[MetaItem]:
        return RepeatedRawMetaValuesView(self)

    def items(self) -> ItemsView[str, MetaItem]:
        return RepeatedRawMetaItemsView(self)


class RepeatedMetaKeysView(_DictView, KeysView[str]):
    def __iter__(self) -> Iterator[str]:
        for item in self._wrapper:
            yield item.key

    def __reversed__(self) -> Iterator[str]:
        for item in reversed(self._wrapper):
            yield item.key


class RepeatedMetaValuesView(_DictView, ValuesView[Optional[MetaValue]]):
    def __iter__(self) -> Iterator[Optional[MetaValue]]:
        for item in self._wrapper:
            yield item.value

    def __reversed__(self) -> Iterator[Optional[MetaValue]]:
        for item in reversed(self._wrapper):
            yield item.value


class RepeatedMetaItemsView(_DictView, ItemsView[str, Optional[MetaValue]]):
    def __iter__(self) -> Iterator[tuple[str, Optional[MetaValue]]]:
        for item in self._wrapper:
            yield item.key, item.value

    def __reversed__(self) -> Iterator[tuple[str, Optional[MetaValue]]]:
        for item in reversed(self._wrapper):
            yield item.key, item.value


class RepeatedMetaItemWrapper(
        internal.RepeatedFilteredNodeWrapper[MetaItem],
        MutableMapping[str, Optional[MetaValue | MetaRawValue]]):

    def __init__(
            self,
            raw_wrapper: internal.RepeatedNodeWithInterleavingCommentsWrapper[MetaItem],
    ) -> None:
        super().__init__(raw_wrapper, MetaItem)

    @overload
    def __getitem__(self, index: int) -> MetaItem:
        ...
    @overload
    def __getitem__(self, index: slice) -> list[MetaItem]:
        ...
    @overload
    def __getitem__(self, index: str) -> Optional[MetaValue]:
        ...
    def __getitem__(self, index: int | slice | str) -> MetaItem | list[MetaItem] | Optional[MetaValue]:
        if not isinstance(index, str):
            return super().__getitem__(index)
        for item in self:
            if item.key == index:
                return item.value
        raise KeyError(index)

    def __delitem__(self, index: int | slice | str) -> None:
        if not isinstance(index, str):
            return super().__delitem__(index)
        for i, item in enumerate(self):
            if item.key == index:
                return super().__delitem__(i)
        raise KeyError(index)

    @overload
    def __setitem__(self, index: int, value: MetaItem) -> None:
        ...
    @overload
    def __setitem__(self, index: slice, value: Iterable[MetaItem]) -> None:
        ...
    @overload
    def __setitem__(self, index: str, value: Optional[MetaValue | MetaRawValue]) -> None:
        ...
    @no_type_check  # mypy does not seem to work well with overload
    def __setitem__(self, index: int | slice | str, value: MetaItem | Iterable[MetaItem] | Optional[MetaValue | MetaRawValue]) -> None:
        if not isinstance(index, str):
            return super().__setitem__(index, value)
        for _, item in enumerate(self):
            if item.key == index:
                item.value = value
                return
        self.append(MetaItem.from_value(index, value, indent=self.indent))

    @overload
    def pop(self, index: int = -1) -> MetaItem:
        ...
    @overload
    def pop(self, index: str) -> Optional[MetaValue]:
        ...
    @overload
    def pop(self, index: str, default: _V) -> Optional[MetaValue] | _V:
        ...
    def pop(self, index: int | str = -1, default: _V | _Empty = _EMPTY) -> MetaItem | Optional[MetaValue] | _V:
        if not isinstance(index, str):
            return super().pop(index)
        for i, item in enumerate(self):
            if item.key != index:
                continue
            item = super().pop(i)
            value = item.value
            if isinstance(value, base.RawModel) and value.token_store:
                if prev := value.token_store.get_prev(value.first_token):
                    value.token_store.remove(item.first_token, prev)
                if next := value.token_store.get_next(value.last_token):
                    value.token_store.remove(next, item.last_token)
            return value
        if not isinstance(default, _Empty):
            return default
        raise KeyError(index)

    def keys(self) -> KeysView[str]:
        return RepeatedMetaKeysView(self)

    def values(self) -> ValuesView[Optional[MetaValue]]:
        return RepeatedMetaValuesView(self)

    def items(self) -> ItemsView[str, Optional[MetaValue]]:
        return RepeatedMetaItemsView(self)


class repeated_raw_meta_item_property(internal.cached_custom_property[RepeatedRawMetaItemWrapper, base.RawTreeModel]):
    def __init__(self, inner_property: internal.repeated_node_with_interleaving_comments_property[MetaItem]):
        super().__init__(
            lambda instance: RepeatedRawMetaItemWrapper(inner_property.__get__(instance)))


class repeated_meta_item_property(internal.cached_custom_property[RepeatedMetaItemWrapper, base.RawTreeModel]):
    def __init__(self, inner_property: internal.repeated_node_with_interleaving_comments_property[MetaItem]):
        super().__init__(
            lambda instance: RepeatedMetaItemWrapper(inner_property.__get__(instance)))


def from_mapping(mapping: Mapping[str, MetaValue | MetaRawValue]) -> Iterator[MetaItem]:
    for key, value in mapping.items():
        yield MetaItem.from_value(key, value)



from typing import Type, TypeVar
from .. import base

_TT = TypeVar('_TT', bound=Type[base.RawTokenModel])
_UT = TypeVar('_UT', bound=Type[base.RawTreeModel])


TOKEN_MODELS: dict[str, Type[base.RawTokenModel]] = {}
TREE_MODELS: dict[str, Type[base.RawTreeModel]] = {}


def token_model(cls: _TT) -> _TT:
    TOKEN_MODELS[cls.RULE] = cls
    return cls


def tree_model(cls: _UT) -> _UT:
    TREE_MODELS[cls.RULE] = cls
    return cls



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import datetime
from typing import Iterable, Mapping, Optional, Type, TypeVar, final
from .. import base, internal, meta_item_internal
from ..amount import Amount
from ..block_comment import BlockComment
from ..currency import Currency
from ..date import Date
from ..inline_comment import InlineComment
from ..meta_item import MetaItem
from ..meta_value import MetaRawValue, MetaValue
from ..punctuation import DedentMark, Eol
from ..spacing import Newline, Whitespace

_Self = TypeVar('_Self', bound='Price')


@internal.token_model
class PriceLabel(internal.SimpleDefaultRawTokenModel):
    RULE = 'PRICE'
    DEFAULT = 'price'


@internal.tree_model
class Price(internal.SurroundingCommentsMixin, base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'price'

    _date = internal.required_field[Date]()
    _label = internal.required_field[PriceLabel]()
    _currency = internal.required_field[Currency]()
    _amount = internal.required_field[Amount]()
    _inline_comment = internal.optional_left_field[InlineComment](separators=(Whitespace.from_default(),))
    _eol = internal.required_field[Eol]()
    _meta = internal.repeated_field[MetaItem | BlockComment](separators=(Newline.from_default(),), default_indent='    ')
    _dedent_mark = internal.optional_left_field[DedentMark](separators=())

    @internal.custom_property
    def _leading_comment_pivot(self) -> base.RawTokenModel:
        return self._date.first_token

    @internal.custom_property
    def _inline_comment_pivot(self) -> base.RawTokenModel:
        return self._amount.last_token

    @internal.custom_property
    def _dedent_mark_pivot(self) -> base.RawTokenModel:
        return self._meta.last_token or self._eol.last_token

    @internal.custom_property
    def _trailing_comment_pivot(self) -> base.RawTokenModel:
        return (self._dedent_mark and self._dedent_mark.last_token) or self._meta.last_token or self._eol.last_token

    raw_leading_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._leading_comment, _leading_comment_pivot)
    raw_date = internal.required_node_property(_date)
    raw_currency = internal.required_node_property(_currency)
    raw_amount = internal.required_node_property(_amount)
    raw_inline_comment = internal.optional_node_property(_inline_comment, _inline_comment_pivot)
    raw_meta_with_comments = internal.repeated_node_with_interleaving_comments_property(_meta)
    raw_meta = meta_item_internal.repeated_raw_meta_item_property(raw_meta_with_comments)
    raw_trailing_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._trailing_comment, _trailing_comment_pivot)

    leading_comment = internal.optional_string_property(raw_leading_comment, BlockComment)
    date = internal.required_value_property(raw_date)
    currency = internal.required_value_property(raw_currency)
    amount = raw_amount
    inline_comment = internal.optional_string_property(raw_inline_comment, InlineComment)
    meta = meta_item_internal.repeated_meta_item_property(raw_meta_with_comments)
    trailing_comment = internal.optional_string_property(raw_trailing_comment, BlockComment)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            leading_comment: Optional[BlockComment],
            date: Date,
            label: PriceLabel,
            currency: Currency,
            amount: Amount,
            inline_comment: Optional[InlineComment],
            eol: Eol,
            repeated_meta: internal.Repeated[MetaItem | BlockComment],
            dedent_mark: Optional[DedentMark],
            trailing_comment: Optional[BlockComment],
    ):
        super().__init__(token_store)
        self._leading_comment = leading_comment
        self._date = date
        self._label = label
        self._currency = currency
        self._amount = amount
        self._inline_comment = inline_comment
        self._eol = eol
        self._meta = repeated_meta
        self._dedent_mark = dedent_mark
        self._trailing_comment = trailing_comment

    @property
    def first_token(self) -> base.RawTokenModel:
        return (self._leading_comment and self._leading_comment.first_token) or self._date.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return (self._trailing_comment and self._trailing_comment.last_token) or (self._dedent_mark and self._dedent_mark.last_token) or self._meta.last_token or self._eol.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._leading_comment.clone(self._leading_comment, token_store, token_transformer),
            type(self)._date.clone(self._date, token_store, token_transformer),
            type(self)._label.clone(self._label, token_store, token_transformer),
            type(self)._currency.clone(self._currency, token_store, token_transformer),
            type(self)._amount.clone(self._amount, token_store, token_transformer),
            type(self)._inline_comment.clone(self._inline_comment, token_store, token_transformer),
            type(self)._eol.clone(self._eol, token_store, token_transformer),
            type(self)._meta.clone(self._meta, token_store, token_transformer),
            type(self)._dedent_mark.clone(self._dedent_mark, token_store, token_transformer),
            type(self)._trailing_comment.clone(self._trailing_comment, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._leading_comment = type(self)._leading_comment.reattach(self._leading_comment, token_store, token_transformer)
        self._date = type(self)._date.reattach(self._date, token_store, token_transformer)
        self._label = type(self)._label.reattach(self._label, token_store, token_transformer)
        self._currency = type(self)._currency.reattach(self._currency, token_store, token_transformer)
        self._amount = type(self)._amount.reattach(self._amount, token_store, token_transformer)
        self._inline_comment = type(self)._inline_comment.reattach(self._inline_comment, token_store, token_transformer)
        self._eol = type(self)._eol.reattach(self._eol, token_store, token_transformer)
        self._meta = type(self)._meta.reattach(self._meta, token_store, token_transformer)
        self._dedent_mark = type(self)._dedent_mark.reattach(self._dedent_mark, token_store, token_transformer)
        self._trailing_comment = type(self)._trailing_comment.reattach(self._trailing_comment, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, Price)
            and self._leading_comment == other._leading_comment
            and self._date == other._date
            and self._label == other._label
            and self._currency == other._currency
            and self._amount == other._amount
            and self._inline_comment == other._inline_comment
            and self._eol == other._eol
            and self._meta == other._meta
            and self._dedent_mark == other._dedent_mark
            and self._trailing_comment == other._trailing_comment
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            date: Date,
            currency: Currency,
            amount: Amount,
            *,
            leading_comment: Optional[BlockComment] = None,
            inline_comment: Optional[InlineComment] = None,
            meta: Iterable[MetaItem | BlockComment] = (),
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        label = PriceLabel.from_default()
        eol = Eol.from_default()
        repeated_meta = cls._meta.create_repeated(meta)
        dedent_mark = None
        tokens = [
            *cls._leading_comment.detach_with_separators(leading_comment),
            *date.detach(),
            Whitespace.from_default(),
            *label.detach(),
            Whitespace.from_default(),
            *currency.detach(),
            Whitespace.from_default(),
            *amount.detach(),
            *cls._inline_comment.detach_with_separators(inline_comment),
            *eol.detach(),
            *cls._meta.detach_with_separators(repeated_meta),
            *cls._dedent_mark.detach_with_separators(dedent_mark),
            *cls._trailing_comment.detach_with_separators(trailing_comment),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._leading_comment.reattach(leading_comment, token_store)
        cls._date.reattach(date, token_store)
        cls._label.reattach(label, token_store)
        cls._currency.reattach(currency, token_store)
        cls._amount.reattach(amount, token_store)
        cls._inline_comment.reattach(inline_comment, token_store)
        cls._eol.reattach(eol, token_store)
        cls._meta.reattach(repeated_meta, token_store)
        cls._dedent_mark.reattach(dedent_mark, token_store)
        cls._trailing_comment.reattach(trailing_comment, token_store)
        return cls(token_store, leading_comment, date, label, currency, amount, inline_comment, eol, repeated_meta, dedent_mark, trailing_comment)

    @classmethod
    def from_value(
            cls: Type[_Self],
            date: datetime.date,
            currency: str,
            amount: Amount,
            *,
            leading_comment: Optional[str] = None,
            inline_comment: Optional[str] = None,
            meta: Optional[Mapping[str, MetaValue | MetaRawValue]] = None,
            trailing_comment: Optional[str] = None,
    ) -> _Self:
        return cls.from_children(
            leading_comment=BlockComment.from_value(leading_comment) if leading_comment is not None else None,
            date=Date.from_value(date),
            currency=Currency.from_value(currency),
            amount=amount,
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            meta=meta_item_internal.from_mapping(meta) if meta is not None else (),
            trailing_comment=BlockComment.from_value(trailing_comment) if trailing_comment is not None else None,
        )

    def auto_claim_comments(self) -> None:
        self.claim_leading_comment(ignore_if_already_claimed=True)
        self.claim_trailing_comment(ignore_if_already_claimed=True)
        type(self)._trailing_comment.auto_claim_comments(self._trailing_comment)
        self.raw_meta_with_comments.auto_claim_comments()
        type(self)._inline_comment.auto_claim_comments(self._inline_comment)
        type(self)._amount.auto_claim_comments(self._amount)
        type(self)._currency.auto_claim_comments(self._currency)
        type(self)._date.auto_claim_comments(self._date)
        type(self)._leading_comment.auto_claim_comments(self._leading_comment)



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import TYPE_CHECKING, Type, TypeVar, final
from .. import base, internal
from ..spacing import Whitespace
if TYPE_CHECKING:
  from ..number_atom_expr import NumberAtomExpr

_Self = TypeVar('_Self', bound='NumberUnaryExpr')


@internal.token_model
class UnaryOp(internal.SimpleRawTokenModel):
    RULE = 'UNARY_OP'


@internal.tree_model
class NumberUnaryExpr(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'number_unary_expr'

    _unary_op = internal.required_field[UnaryOp]()
    _operand = internal.required_field['NumberAtomExpr']()

    raw_unary_op = internal.required_node_property(_unary_op)
    raw_operand = internal.required_node_property(_operand)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            unary_op: UnaryOp,
            operand: 'NumberAtomExpr',
    ):
        super().__init__(token_store)
        self._unary_op = unary_op
        self._operand = operand

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._unary_op.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._operand.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._unary_op.clone(self._unary_op, token_store, token_transformer),
            type(self)._operand.clone(self._operand, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._unary_op = type(self)._unary_op.reattach(self._unary_op, token_store, token_transformer)
        self._operand = type(self)._operand.reattach(self._operand, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, NumberUnaryExpr)
            and self._unary_op == other._unary_op
            and self._operand == other._operand
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            unary_op: UnaryOp,
            operand: 'NumberAtomExpr',
    ) -> _Self:
        tokens = [
            *unary_op.detach(),
            Whitespace.from_default(),
            *operand.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._unary_op.reattach(unary_op, token_store)
        cls._operand.reattach(operand, token_store)
        return cls(token_store, unary_op, operand)

    def auto_claim_comments(self) -> None:
        type(self)._operand.auto_claim_comments(self._operand)
        type(self)._unary_op.auto_claim_comments(self._unary_op)



import abc
import copy
from typing import Any, ClassVar, Optional, Type, TypeVar, cast
from autobean_refactor import token_store as token_store_lib

_T = TypeVar('_T', bound='RawTokenModel')
# TODO: replace with PEP 673 Self once supported
_SelfRawModel = TypeVar('_SelfRawModel', bound='RawModel')
_SelfRawTokenModel = TypeVar('_SelfRawTokenModel', bound='RawTokenModel')  
_SelfRawTreeModel = TypeVar('_SelfRawTreeModel', bound='RawTreeModel')
TokenStore = token_store_lib.TokenStore['RawTokenModel']


# This could technically be replaced with map.__getitem__ but that didn't work well with mypy.
# https://github.com/python/mypy/issues/1317
class TokenTransformer(abc.ABC):
    @abc.abstractmethod
    def transform(self, token: _T) -> _T:
        ...


class MappingTokenTransformer(TokenTransformer):
    def __init__(self, map: dict[int, 'RawTokenModel']) -> None:
        self._map = map

    def transform(self, token: _T) -> _T:
        return cast(_T, self._map[id(token)])


class IdentityTokenTransformer(TokenTransformer):
    def transform(self, token: _T) -> _T:
        return token


IDENTITY_TOKEN_TRANSFORMER = IdentityTokenTransformer()


class RawModel(abc.ABC):
    RULE: ClassVar[str]

    @property
    @abc.abstractmethod
    def token_store(self) -> Optional[TokenStore]:
        ...

    @property
    @abc.abstractmethod
    def first_token(self) -> 'RawTokenModel':
        ...

    @property
    @abc.abstractmethod
    def last_token(self) -> 'RawTokenModel':
        ...

    @abc.abstractmethod
    def __deepcopy__(self: _SelfRawModel, memo: dict[int, Any]) -> _SelfRawModel:
        ...

    def detach(self) -> list['RawTokenModel']:
        if not self.token_store:
            return []
        if (
                self.first_token is not self.token_store.get_first() or
                self.last_token is not self.token_store.get_last()):
            raise ValueError('Cannot reuse node. Consider making a copy.')
        tokens = list(self.token_store)
        if tokens:
            self.token_store.remove(tokens[0], tokens[-1])
        return tokens

    @property
    def tokens(self) -> list['RawTokenModel']:
        if not self.token_store or not self.first_token or not self.last_token:
            return []
        return list(self.token_store.iter(self.first_token, self.last_token))

    @abc.abstractmethod
    def __eq__(self, other: object) -> bool:
        ...

    @abc.abstractmethod
    def clone(self: _SelfRawModel, token_store: TokenStore, token_transformer: TokenTransformer) -> _SelfRawModel:
        ...

    @abc.abstractmethod
    def reattach(self: _SelfRawModel, token_store: TokenStore, token_transformer: TokenTransformer = IDENTITY_TOKEN_TRANSFORMER) -> _SelfRawModel:
        ...

    @abc.abstractmethod
    def auto_claim_comments(self) -> None:
        ...


class RawTokenModel(token_store_lib.Token, RawModel):
    def __init__(self, raw_text: str) -> None:
        super().__init__(raw_text)

    @classmethod
    def from_raw_text(cls: Type[_SelfRawTokenModel], raw_text: str) -> _SelfRawTokenModel:
        return cls(raw_text)

    @property
    def token_store(self) -> Optional[TokenStore]:
        return self.store_handle.store if self.store_handle else None

    @property
    def first_token(self) -> 'RawTokenModel':
        return self

    @property
    def last_token(self) -> 'RawTokenModel':
        return self

    @abc.abstractmethod
    def _clone(self: _SelfRawTokenModel) -> _SelfRawTokenModel:
        ...

    def __deepcopy__(self: _SelfRawTokenModel, memo: dict[int, Any]) -> _SelfRawTokenModel:
        del memo  # unused
        return self._clone()

    def detach(self) -> list['RawTokenModel']:
        if not self.store_handle:
            return [self]
        return super().detach()

    @property
    def tokens(self) -> list['RawTokenModel']:
        return [self]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RawTokenModel) and self.RULE == other.RULE and self.raw_text == other.raw_text

    def clone(self: _SelfRawTokenModel, token_store: TokenStore, token_transformer: TokenTransformer) -> _SelfRawTokenModel:
        return token_transformer.transform(self)

    def reattach(self: _SelfRawTokenModel, token_store: TokenStore, token_transformer: TokenTransformer = IDENTITY_TOKEN_TRANSFORMER) -> _SelfRawTokenModel:
        return token_transformer.transform(self)

    def auto_claim_comments(self) -> None:
        pass

    def __hash__(self) -> int:
        return hash((type(self).RULE, self.raw_text))


class RawTreeModel(RawModel):
    def __init__(self, token_store: TokenStore) -> None:
        super().__init__()
        self._token_store = token_store

    @classmethod
    def from_parsed_children(cls: Type[_SelfRawTreeModel], token_store: TokenStore, *children: Optional[RawModel]) -> _SelfRawTreeModel:
        return cls(token_store, *children)

    @property
    def token_store(self) -> TokenStore:
        return self._token_store

    def __deepcopy__(self: _SelfRawTreeModel, memo: dict[int, Any]) -> _SelfRawTreeModel:
        del memo  # unused
        tokens: list[RawTokenModel] = []
        token_map: dict[int, RawTokenModel] = {}
        for token in self._token_store.iter(self.first_token, self.last_token):
            new_token = copy.deepcopy(token)
            tokens.append(new_token)
            token_map[id(token)] = new_token
        token_store = TokenStore.from_tokens(tokens)
        return self.clone(token_store, MappingTokenTransformer(token_map))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RawTreeModel) and self.tokens == other.tokens and self._eq(other)

    @abc.abstractmethod
    def _eq(self, other: 'RawTreeModel') -> bool:
        ...

    def reattach(self: _SelfRawTreeModel, token_store: TokenStore, token_transformer: TokenTransformer = IDENTITY_TOKEN_TRANSFORMER) -> _SelfRawTreeModel:
        self._reattach(token_store, token_transformer)
        return self

    @abc.abstractmethod
    def _reattach(self, token_store: TokenStore, token_transformer: TokenTransformer) -> None:
        ...



def range_from_index(index: int | slice, length: int) -> range:
    if isinstance(index, int):
        try:
            index = range(length)[index]
        except IndexError:
            raise IndexError('list assignment index out of range')
        return range(index, index + 1)
    return range(length)[index]


def slice_from_range(r: range) -> slice:
    stop = r.stop if r.stop != -1 else None
    return slice(r.start, stop, r.step)



import decimal
from typing import Any, Optional, Type, TypeVar, cast, final, TYPE_CHECKING
from . import base
from . import internal
if TYPE_CHECKING:
    from .number_mul_expr import NumberMulExpr
else:
    NumberMulExpr = Any

# TODO: replace with PEP 673 Self once supported
_Self = TypeVar('_Self', bound='NumberAddExpr')


@internal.token_model
class AddOp(internal.SimpleRawTokenModel):
    RULE = 'ADD_OP'


@internal.tree_model
class NumberAddExpr(base.RawTreeModel):
    RULE = 'number_add_expr'

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            operands: tuple[NumberMulExpr, ...],
            ops: tuple[AddOp, ...],
    ):
        super().__init__(token_store)
        self._raw_operands = operands
        self._raw_ops = ops

    @classmethod
    def from_parsed_children(cls: Type[_Self], token_store: base.TokenStore, *children: Optional[base.RawModel]) -> _Self:
        return cls(
            token_store,
            cast(tuple[NumberMulExpr, ...], children[::2]),
            cast(tuple[AddOp, ...], children[1::2]))

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._raw_operands[0].first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._raw_operands[-1].last_token

    @property
    def raw_operands(self) -> tuple[NumberMulExpr, ...]:
        return self._raw_operands

    @property
    def raw_ops(self) -> tuple[AddOp, ...]:
        return self._raw_ops

    @property
    def value(self) -> decimal.Decimal:
        value = self._raw_operands[0].value
        for op, operand in zip(self._raw_ops, self._raw_operands[1:]):
            if op.raw_text == '+':
                value += operand.value
            elif op.raw_text == '-':
                value -= operand.value
            else:
                assert False
        return value

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        ops = tuple(op.clone(token_store, token_transformer) for op in self._raw_ops)
        operands = tuple(operand.clone(token_store, token_transformer) for operand in self._raw_operands)
        return type(self)(token_store, operands, ops)

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._raw_ops = tuple(op.reattach(token_store, token_transformer) for op in self._raw_ops)
        self._raw_operands = tuple(operand.reattach(token_store, token_transformer) for operand in self._raw_operands)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, NumberAddExpr)
            and self._raw_operands == other._raw_operands
            and self._raw_ops == other._raw_ops)

    def auto_claim_comments(self) -> None:
        pass  # no block comments



from typing import TypeVar
from . import base_token_models, spacing_accessors

_V = TypeVar('_V')


class SimpleRawTokenModel(
        base_token_models.SimpleRawTokenModel,
        spacing_accessors.SpacingAccessorsMixin):
    pass


class SingleValueRawTokenModel(
        base_token_models.SingleValueRawTokenModel[_V],
        spacing_accessors.SpacingAccessorsMixin):
    pass


class SimpleSingleValueRawTokenModel(
        base_token_models.SimpleSingleValueRawTokenModel,
        spacing_accessors.SpacingAccessorsMixin):
    pass


class DefaultRawTokenModel(
        base_token_models.DefaultRawTokenModel,
        spacing_accessors.SpacingAccessorsMixin):
    pass


class SimpleDefaultRawTokenModel(
        base_token_models.SimpleDefaultRawTokenModel,
        spacing_accessors.SpacingAccessorsMixin):
    pass



from . import internal


@internal.token_model
class Tag(internal.SingleValueRawTokenModel[str]):
    RULE = 'TAG'

    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        return raw_text[1:]

    @classmethod
    def _format_value(cls, value: str) -> str:
        return f'#{value}'



from .date import Date
from .punctuation import Asterisk
from .escaped_string import EscapedString
from .currency import Currency
from .number_expr import NumberExpr
from .amount import Amount
from .compound_amount import CompoundAmount

CostComponent = Date | Asterisk | EscapedString | Currency | NumberExpr | Amount | CompoundAmount



from . import internal
from .generated import unit_cost, total_cost
from .generated.unit_cost import LeftBrace, RightBrace
from .generated.total_cost import DblLeftBrace, DblRightBrace


@internal.tree_model
class UnitCost(unit_cost.UnitCost):
    def into_total_cost(self) -> 'TotalCost':
        dbl_left_brace = DblLeftBrace.from_default()
        dbl_right_brace = DblRightBrace.from_default()
        self.token_store.replace(self._left_brace, dbl_left_brace)
        self.token_store.replace(self._right_brace, dbl_right_brace)
        return TotalCost(self.token_store, dbl_left_brace, self._components, dbl_right_brace)


@internal.tree_model
class TotalCost(total_cost.TotalCost):
    def into_unit_cost(self) -> UnitCost:
        left_brace = LeftBrace.from_default()
        right_brace = RightBrace.from_default()
        self.token_store.replace(self._dbl_left_brace, left_brace)
        self.token_store.replace(self._dbl_right_brace, right_brace)
        return UnitCost(self.token_store, left_brace, self._components, right_brace)



import collections
import copy
import enum
import glob
import os.path
import pathlib
import re
from typing import Iterable, Iterator, Optional, Type, TypeVar
import lark
from lark import exceptions
from lark import lexer
from lark import load_grammar
from autobean_refactor import models
from autobean_refactor.models import internal

_T = TypeVar('_T', bound=models.RawTokenModel)
_U = TypeVar('_U', bound=models.RawTreeModel)


with open(pathlib.Path(__file__).parent / 'beancount.lark') as f:
    _GRAMMAR, _ = load_grammar.load_grammar(
        grammar=f.read(),
        source=f.name,
        import_paths=[],
        global_keep_all_tokens=False)
    _IGNORED_TOKENS = frozenset(_GRAMMAR.ignore)
    _GRAMMAR.ignore.clear()


class PostLexInline(lark.lark.PostLex):
    always_accept = _IGNORED_TOKENS

    def process(self, stream: Iterator[lark.Token]) -> Iterator[lark.Token]:
        return stream


class PostLex(lark.lark.PostLex):
    _NEWLINE_INDENT_COMMENT_SPLIT_RE = re.compile(r'([\r\n]*)([ \t]*)(;.*)?', re.S)
    _NEWLINE_INDENT_COMMENT = '_NEWLINE_INDENT_COMMENT'
    _NEWLINE = '_NEWLINE'
    _EOL = 'EOL'
    _INDENT_MARK = 'INDENT_MARK'
    _DEDENT_MARK = 'DEDENT_MARK'
    _INDENT = 'INDENT'
    _BLOCK_COMMENT = 'BLOCK_COMMENT'

    # Contextual lexer only sees _EOL and will thus reject _NEWLINE by default.
    always_accept = _IGNORED_TOKENS | {_NEWLINE_INDENT_COMMENT}

    def process(self, stream: Iterator[lark.Token]) -> Iterator[lark.Token]:
        indented = False
        token = None
        prev_is_block_comment = False
        for token in stream:
            if token.type != self._NEWLINE_INDENT_COMMENT:
                yield token
                continue
            match = self._NEWLINE_INDENT_COMMENT_SPLIT_RE.fullmatch(token.value)
            assert match
            newline_text, indent_text, comment_text = match.groups()
            if newline_text and not prev_is_block_comment:
                yield lark.Token.new_borrow_pos(self._EOL, '', token)
            if not indent_text and indented:
                indented = False
                yield lark.Token.new_borrow_pos(self._DEDENT_MARK, '', token)
            if newline_text:
                yield lark.Token.new_borrow_pos(self._NEWLINE, newline_text, token)
            prev_is_block_comment = False
            if indent_text and not indented:
                indented = True
                yield lark.Token.new_borrow_pos(self._INDENT_MARK, '', token)
            if comment_text:
                prev_is_block_comment = True
                yield lark.Token.new_borrow_pos(self._BLOCK_COMMENT, indent_text + comment_text, token)
            elif indent_text:
                yield lark.Token.new_borrow_pos(self._INDENT, indent_text, token)
        if not prev_is_block_comment:
            yield lark.Token(self._EOL, '')
        if indented:
            yield lark.Token(self._DEDENT_MARK, '')


class _Floating(enum.Enum):
    LEFT = enum.auto()
    RIGHT = enum.auto()


def _get_include_paths(path: str, file: models.File) -> Iterable[str]:
    for directive in file.raw_directives:
        if not isinstance(directive, models.Include):
            continue
        matches = glob.glob(os.path.join(os.path.dirname(path), directive.filename), recursive=True)
        if not matches:
            lineno = directive.token_store.get_position(directive.first_token).line
            raise ValueError(f'No files match {directive.filename!r} ({path}:{lineno})')
        yield from matches


class Parser:
    _lark: lark.Lark
    _token_models: dict[str, Type[models.RawTokenModel]]
    _tree_models: dict[str, Type[models.RawTreeModel]]

    def __init__(self) -> None:
        start = list(models.TREE_MODELS.keys())
        self._lark = lark.Lark(
            _GRAMMAR, lexer='contextual', parser='lalr', postlex=PostLex(), start=start)
        self._lark_inline = lark.Lark(
            _GRAMMAR, lexer='contextual', parser='lalr', postlex=PostLexInline(), start=start)

    def parse_token(self, text: str, target: Type[_T]) -> _T:
        """Parses one token.

        This is a separate method to ease typing and support ignored tokens.
        """
        lexer_conf = copy.deepcopy(self._lark.parser.lexer_conf)
        lexer_conf.terminals = [lexer_conf.terminals_by_name[target.RULE]]
        basic_lexer = lexer.BasicLexer(lexer_conf)
        lexer_thread = lexer.LexerThread.from_text(basic_lexer, text)
        tokens = list(lexer_thread.lex(None))
        if not tokens:
            raise exceptions.UnexpectedToken(
                lark.Token('$END', '', 0, 1, 1), {target.RULE})
        if tokens[0].type != target.RULE:
            raise exceptions.UnexpectedToken(tokens[0], {target.RULE})
        if len(tokens) > 1:
            raise exceptions.UnexpectedToken(tokens[1], {'$END'})
        return target.from_raw_text(tokens[0].value)

    def parse_inline(self, text: str, target: Type[_U]) -> _U:
        return self._parse(text, target, self._lark_inline)

    def parse(self, text: str, target: Type[_U]) -> _U:
        return self._parse(text, target, self._lark)

    def parse_file_recursive(self, path: str) -> dict[str, models.File]:
        files = {}
        queue = collections.deque([path])

        while queue:
            path = queue.popleft()
            if path in files:
                continue
            with open(path) as f:
                text = f.read()
            file = self.parse(text, models.File)
            files[path] = file
            queue.extend(_get_include_paths(path, file))
        
        return files

    def _parse(self, text: str, target: Type[_U], lark_instance: lark.Lark) -> _U:
        parser = lark_instance.parse_interactive(text=text, start=target.RULE)
        tokens = []
        for token in parser.lexer_thread.lex(parser.parser_state):
            tokens.append(token)
            if token.type not in _IGNORED_TOKENS or token.type in parser.choices():
                parser.feed_token(token)
        tree = parser.feed_eof()

        return ModelBuilder(tokens).build(tree, target)


class ModelBuilder:
    def __init__(self, tokens: list[lark.Token]):
        self._tokens = tokens
        self._built_tokens: list[models.RawTokenModel] = []
        self._token_to_index = {id(token): i for i, token in enumerate(tokens)}
        self._cursor = 0
        self._token_store = models.TokenStore.from_tokens([])

    def _add_tokens(self, tokens: Iterable[models.RawTokenModel]) -> None:
        self._built_tokens.extend(tokens)

    def _fix_gap(self, cursor: int) -> None:
        for token in self._tokens[self._cursor:cursor]:
            if not token.value:  # skips EOL, INDENT_MARK, DEDENT_MARK, etc. if outside a model.
                continue
            built_token = models.TOKEN_MODELS[token.type].from_raw_text(token.value)
            if isinstance(built_token, models.BlockComment):
                built_token.claimed = False
            self._add_tokens([built_token])
        self._cursor = cursor

    def _build_indent(self) -> models.RawTokenModel:
        cursor = self._cursor
        while cursor < len(self._tokens):
            token = self._tokens[cursor]
            if token.value:
                if token.type == 'INDENT':
                    self._fix_gap(cursor)
                    return self._build_token(token)
                if not token.type in _IGNORED_TOKENS:
                    break
            cursor += 1
        raise exceptions.UnexpectedInput('Missing indent.')

    def _build_placeholder(self, floating: _Floating) -> internal.Placeholder:
        placeholder = internal.Placeholder.from_default()
        if floating == _Floating.RIGHT:
            assert False, 'deprecated code path'
        elif floating == _Floating.LEFT:
            self._built_tokens.append(placeholder)
        else:
            assert False
        return placeholder

    def _build_token(self, token: lark.Token) -> models.RawTokenModel:
        self._fix_gap(self._token_to_index[id(token)])
        built_token = models.TOKEN_MODELS[token.type].from_raw_text(token.value)
        self._add_tokens([built_token])
        self._cursor += 1
        return built_token

    def _build_tree(self, tree: lark.Tree) -> models.RawTreeModel:
        model_type = models.TREE_MODELS[tree.data]
        children: list[Optional[models.RawModel]] = []
        for child in tree.children:
            is_tree = isinstance(child, lark.Tree)
            if child is None:
                children.append(child)
            elif is_tree and child.data in ('repeated', 'repeated_sep'):
                children.append(self._build_repeated_node(child))
            elif is_tree and child.data in ('indent', 'indent2'):
                children.append(self._build_indent())
            elif is_tree and child.data.endswith('_'):
                continue
            else:
                children.append(self._build_required_node(child))
        return model_type.from_parsed_children(self._token_store, *children)

    def _build_required_node(self, node: lark.Token | lark.Tree) -> models.RawModel:
        if isinstance(node, lark.Token):
            return self._build_token(node)
        if isinstance(node, lark.Tree):
            return self._build_tree(node)
        assert False

    def _build_repeated_node(self, node: lark.Tree) -> models.RawModel:
        placeholder = self._build_placeholder(_Floating.LEFT)
        items = [
            self._build_required_node(child) for child in node.children
            if not (isinstance(child, lark.Tree) and child.data.endswith('_'))
        ]
        indent = getattr(items[0], 'indent', None) if items else None
        return internal.Repeated(self._token_store, items, placeholder, indent)

    def build(self, tree: lark.Tree, model_type: Type[_U]) -> _U:
        model = self._build_tree(tree)
        self._fix_gap(len(self._tokens))
        self._token_store.insert_after(None, self._built_tokens)
        assert isinstance(model, model_type)
        return model



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import Type, TypeVar, final
from .. import base, internal
from ..cost import TotalCost, UnitCost

_Self = TypeVar('_Self', bound='CostSpec')


@internal.tree_model
class CostSpec(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'cost_spec'

    _cost = internal.required_field[TotalCost | UnitCost]()

    raw_cost = internal.required_node_property[TotalCost | UnitCost](_cost)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            cost: TotalCost | UnitCost,
    ):
        super().__init__(token_store)
        self._cost = cost

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._cost.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._cost.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._cost.clone(self._cost, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._cost = type(self)._cost.reattach(self._cost, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, CostSpec)
            and self._cost == other._cost
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            cost: TotalCost | UnitCost,
    ) -> _Self:
        tokens = [
            *cost.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._cost.reattach(cost, token_store)
        return cls(token_store, cost)

    def auto_claim_comments(self) -> None:
        type(self)._cost.auto_claim_comments(self._cost)



import abc
from typing import Generic, Optional, Type, TypeVar, overload
from .. import base

_U = TypeVar('_U', bound=base.RawTreeModel)
_V = TypeVar('_V')
_V_cov = TypeVar('_V_cov', covariant=True)
_Self = TypeVar('_Self', bound='base_ro_property')


class base_ro_property(Generic[_V_cov, _U], abc.ABC):

    @abc.abstractmethod
    def _get(self, instance: _U) -> _V_cov:
        ...

    @overload
    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> _V_cov:
        ...

    @overload
    def __get__(self: _Self, instance: None, owner: Type[_U]) -> _Self:
        ...

    def __get__(self: _Self, instance: Optional[_U], owner: Optional[Type[_U]] = None) -> _V_cov | _Self:
        del owner
        if instance is None:
            return self
        return self._get(instance)


class base_rw_property(base_ro_property[_V, _U]):

    @abc.abstractmethod
    def __set__(self, instance: _U, value: _V) -> None:
        ...



import datetime
import decimal
from typing import Iterable, Iterator, Mapping, NoReturn, Optional, Type, TypeVar, Union, cast, get_args

from . import internal, meta_item_internal
from .generated import custom
from .generated.custom import CustomLabel, CustomRawValue
from .block_comment import BlockComment
from .escaped_string import EscapedString
from .date import Date
from .account import Account
from .amount import Amount
from .bool import Bool
from .inline_comment import InlineComment
from .number_expr import NumberExpr
from .number_unary_expr import NumberUnaryExpr
from .meta_item import MetaItem
from .meta_value import MetaRawValue, MetaValue

_ValueTypeSimplified = str | datetime.date | bool | decimal.Decimal
_ValueTypePreserved = Account | Amount
CustomValue = Union[_ValueTypeSimplified, _ValueTypePreserved]
_Self = TypeVar('_Self', bound='Custom')


def _simplify_value(raw_value: CustomRawValue) -> CustomValue:
    if isinstance(raw_value, EscapedString | Date | Bool | NumberExpr):
        return raw_value.value
    return raw_value


def _unsimplify_value(value: CustomValue | CustomRawValue) -> CustomRawValue:
    match value:
        case str():
            return EscapedString.from_value(value)
        case datetime.date():
            return Date.from_value(value)
        case bool():
            return Bool.from_value(value)
        case decimal.Decimal():
            return NumberExpr.from_value(value)
        case _:
            return value


def _disambiguate_values(values: Iterable[CustomRawValue]) -> Iterator[CustomRawValue]:
    prev = None
    for value in values:
        if isinstance(prev, NumberExpr):
            if isinstance(value, Amount):
                number = value.raw_number
            elif isinstance(value, NumberExpr):
                number = value
            else:
                number = None
            if number is not None and isinstance(
                    number.raw_number_add_expr.raw_operands[0].raw_operands[0], NumberUnaryExpr):
                number.wrap_with_parenthesis()
        yield value
        prev = value


def _update_raw(raw_value: CustomRawValue, value: CustomValue) -> bool:
    match raw_value, value:
        case EscapedString() as r, str() as v:
            r.value = v
        case Date() as r, datetime.date() as v:
            r.value = v
        case Bool() as r, bool() as v:
            r.value = v
        case NumberExpr() as r, decimal.Decimal() as v:
            r.value = v
        case _:
            return False
    return True


# TODO: disambiguate values inserted through repeated wrapper
@internal.tree_model
class Custom(custom.Custom):

    @internal.cached_custom_property
    def values(self) -> internal.RepeatedValueWrapper[CustomRawValue, CustomValue]:
        return internal.RepeatedValueWrapper[CustomRawValue, CustomValue](
            raw_wrapper=self.raw_values,
            raw_type=get_args(CustomRawValue),
            from_raw_type=_simplify_value,
            to_raw_type=_unsimplify_value,
            update_raw=_update_raw,
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            date: Date,
            type: EscapedString,
            values: Iterable[CustomRawValue],
            *,
            leading_comment: Optional[BlockComment] = None,
            inline_comment: Optional[InlineComment] = None,
            meta: Iterable[MetaItem | BlockComment] = (),
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        return super().from_children(
            date,
            type,
            _disambiguate_values(values),
            leading_comment=leading_comment,
            inline_comment=inline_comment,
            meta=meta,
            trailing_comment=trailing_comment)

    @classmethod
    def from_value(
            cls: Type[_Self],
            date: datetime.date,
            type: str,
            values: Iterable[CustomValue | CustomRawValue],
            *,
            inline_comment: Optional[str] = None,
            meta: Optional[Mapping[str, MetaValue | MetaRawValue]] = None,
    ) -> _Self:
        return cls.from_children(
            date=Date.from_value(date),
            type=EscapedString.from_value(type),
            values=map(_unsimplify_value, values),
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            meta=meta_item_internal.from_mapping(meta) if meta is not None else (),
        )



import decimal
from . import internal
from .generated import number_unary_expr
from .generated.number_unary_expr import UnaryOp


@internal.tree_model
class NumberUnaryExpr(number_unary_expr.NumberUnaryExpr):
    @property
    def value(self) -> decimal.Decimal:
        if self._unary_op.raw_text == '+':
            return self._operand.value
        elif self._unary_op.raw_text == '-':
            return -self._operand.value
        else:
            assert False



import abc
import copy
import functools
import itertools
from typing import Any, Callable, Collection, Iterable, MutableSequence, Optional, Type, TypeVar, overload
from .base_property import base_ro_property, base_rw_property
from .fields import required_field, optional_field, repeated_field
from .repeated import Repeated
from . import indexes
from .. import base

_M = TypeVar('_M', bound=base.RawModel)
_U = TypeVar('_U', bound=base.RawTreeModel)
_V = TypeVar('_V')


def replace_node(node: _M, repl: _M) -> None:
    token_store = node.token_store  # backup because the RawTokenModel.token_store may disappear
    if not token_store:
        raise ValueError('Cannot replace a free token.')
    if node is repl:
        return
    token_store.splice(repl.detach(), node.first_token, node.last_token)
    if isinstance(repl, base.RawTreeModel):
        repl.reattach(token_store)


class required_node_property(base_rw_property[_M, base.RawTreeModel]):
    def __init__(self, inner_field: required_field[_M]) -> None:
        super().__init__()
        self._inner_field = inner_field

    def _get(self, instance: base.RawTreeModel) -> _M:
        return self._inner_field.__get__(instance)

    def __set__(self, instance: base.RawTreeModel, value: _M) -> None:
        assert value is not None
        current = self._inner_field.__get__(instance)
        replace_node(current, value)
        self._inner_field.__set__(instance, value)


class optional_node_property(base_rw_property[Optional[_M], _U]):
    def __init__(
            self,
            inner_field: optional_field[_M],
            pivot_property: base_ro_property[base.RawTokenModel, _U],
    ) -> None:
        super().__init__()
        self._inner_field = inner_field
        self._pivot_property = pivot_property

    def _get(self, instance: _U) -> Optional[_M]:
        return self._inner_field.__get__(instance)

    def __set__(self, instance: _U, value: Optional[_M]) -> None:
        current = self._inner_field.__get__(instance)
        if current is None and value is not None:
            pivot = self._pivot_property.__get__(instance)
            self._inner_field._create_node(instance.token_store, pivot, value)
        elif current is not None and value is None:
            pivot = self._pivot_property.__get__(instance)
            self._inner_field._remove_node(instance.token_store, pivot, current)
        elif current is not None and value is not None:
            replace_node(current, value)
        self._inner_field.__set__(instance, value)


class RepeatedNodeWrapper(MutableSequence[_M]):
    def __init__(self, repeated: Repeated[_M], field: repeated_field) -> None:
        self._repeated = repeated
        self._field = field

    @property
    def repeated(self) -> Repeated[_M]:
        return self._repeated

    @functools.cached_property
    def indent(self) -> Optional[str]:
        if self._field.default_indent is None:
            return None
        elif self._repeated.inferred_indent is not None:
            return self._repeated.inferred_indent
        else:
            return self._field.default_indent

    @functools.cached_property
    def _separators(self) -> tuple[base.RawTokenModel, ...]:
        return self._field.separators

    @functools.cached_property
    def _separators_before(self) -> tuple[base.RawTokenModel, ...]:
        return (
            self._field.separators_before
            if self._field.separators_before is not None else self._field.separators)

    def __len__(self) -> int:
        return len(self._repeated.items)

    @overload
    def __getitem__(self, index: int) -> _M:
        ...
    @overload
    def __getitem__(self, index: slice) -> list[_M]:
        ...
    def __getitem__(self, index: int | slice) -> _M | list[_M]:
        return self._repeated.items[index]

    def __delitem__(self, index: int | slice) -> None:
        r = indexes.range_from_index(index, len(self._repeated.items))
        if r.step == 1:
            self[indexes.slice_from_range(r)] = []
        else:
            self.drop_many(r)

    def _insert_tokens(
            self,
            index: int,
            values: Iterable[base.RawModel],
            length: Optional[int] = None,
            separators_before_last: Optional[base.RawTokenModel] = None,
    ) -> None:
        tokens: list[base.RawTokenModel] = []
        ref = self._prev_last(index)
        if length is None:
            length = len(self._repeated.items)
        for i, value in enumerate(values):
            if i or index:
                tokens.extend(copy.deepcopy(self._separators))
                tokens.extend(value.detach())
            elif length:
                tokens.extend(value.detach())
                tokens.extend(copy.deepcopy(self._separators))
                if separators_before_last is None:
                    separators_before_last = self._repeated.token_store.get_prev(
                        self._repeated.items[0].first_token)
                    assert separators_before_last is not None
                ref = separators_before_last
            else:
                tokens.extend(copy.deepcopy(self._separators_before))
                tokens.extend(value.detach())
        self._repeated.token_store.insert_after(ref, tokens)

    def _prev_last(self, index: int) -> base.RawTokenModel:
        return (self._repeated.items[index - 1].last_token
            if index > 0 else self._repeated.placeholder)

    def _del_tokens(self, start: int, stop: int) -> None:
        # Consecutive tokens must be deleted in the same call.
        if stop <= start:
            return
        if start == 0 and stop < len(self._repeated.items):
            first_token = self._repeated.items[start].first_token
            next_first = self._repeated.items[stop].first_token
            t = self._repeated.token_store.get_prev(next_first)
            assert t is not None
            last_token = t
        else:
            prev_last = self._prev_last(start)
            t = self._repeated.token_store.get_next(prev_last)
            assert t is not None
            first_token = t
            last_token = self._repeated.items[stop - 1].last_token
        self._repeated.token_store.remove(first_token, last_token)

    @overload
    def __setitem__(self, index: int, value: _M) -> None:
        ...
    @overload
    def __setitem__(self, index: slice, value: Iterable[_M]) -> None:
        ...
    def __setitem__(self, index: int | slice, value: _M | Iterable[_M]) -> None:
        if isinstance(index, int):
            assert not isinstance(value, Iterable)
            item = self._repeated.items[index]
            self._repeated.token_store.splice(value.detach(), item.first_token, item.last_token)
            value.reattach(self._repeated.token_store)
            self._repeated.items[index] = value
            return
        assert isinstance(value, Iterable)
        values = list(value)
        r = indexes.range_from_index(index, len(self._repeated.items))
        separators_before_last = (
            self._repeated.token_store.get_prev(self._repeated.items[0].first_token)
            if self._repeated.items else None)
        if r.step == 1:
            self._del_tokens(r.start, r.stop)
            self._insert_tokens(
                r.start, values, len(self._repeated.items) - len(r), separators_before_last)
            self._repeated.items[indexes.slice_from_range(r)] = values
        else:
            if len(r) != len(values):
                raise ValueError(f'attempt to assign sequence of size {len(values)} to extended slice of size {len(r)}')
            for i, value in zip(r, values):
                self._del_tokens(i, i + 1)
                self._insert_tokens(i, [value], len(self._repeated.items) - 1, separators_before_last)
                value.reattach(self._repeated.token_store)
                self._repeated.items[i] = value

    def insert(self, index: int, value: _M) -> None:
        index = min(index, len(self._repeated.items))
        self._insert_tokens(index, [value])
        value.reattach(self._repeated.token_store)
        self._repeated.items.insert(index, value)

    def append(self, value: _M) -> None:
        self._insert_tokens(len(self._repeated.items), [value])
        value.reattach(self._repeated.token_store)
        self._repeated.items.append(value)

    def clear(self) -> None:
        self._del_tokens(0, len(self._repeated.items))
        self._repeated.items.clear()

    def extend(self, values: Iterable[_M]) -> None:
        values = list(values)
        self._insert_tokens(len(self._repeated.items), values)
        self._repeated.items.extend(values)

    def pop(self, index: int = -1) -> _M:
        value = self._repeated.items[index]
        tokens = value.tokens
        r = indexes.range_from_index(index, len(self._repeated.items))
        self._del_tokens(r.start, r.stop)
        token_store = base.TokenStore.from_tokens(tokens)
        value.reattach(token_store)
        self._repeated.items.pop(index)
        return value

    def __deepcopy__(self, memo: dict[int, Any]) -> 'RepeatedNodeWrapper':
        repeated = copy.deepcopy(self._repeated, memo)
        return RepeatedNodeWrapper(repeated, self._field)

    def drop_many(self, indexes: Iterable[int]) -> None:
        indexes = sorted(indexes, reverse=True)
        count = itertools.count()
        ranges = (
            list(r) for _, r in itertools.groupby(indexes, key=lambda i: i + next(count))
        )
        for r in ranges:
            self._del_tokens(r[-1], r[0] + 1)
        self._repeated.items[:] = (
            item for i, item in enumerate(self._repeated.items) if i not in indexes
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Collection) and
            all(a == b for a, b in itertools.zip_longest(self, other)))

    def auto_claim_comments(self) -> None:
        self._repeated.auto_claim_comments()


class repeated_node_property(base_rw_property[RepeatedNodeWrapper[_M], base.RawTreeModel]):
    def __init__(self, inner_field: repeated_field[_M]) -> None:
        super().__init__()
        self._inner_field = inner_field

    def __set_name__(self, owner: base.RawTreeModel, name: str) -> None:
        self._attr = name

    def _get(self, instance: _U) -> RepeatedNodeWrapper[_M]:
        wrapper = instance.__dict__.get(self._attr)
        if wrapper is None:
            repeated = self._inner_field.__get__(instance)
            wrapper = RepeatedNodeWrapper(repeated, self._inner_field)
            instance.__dict__[self._attr] = wrapper
        return wrapper

    def __set__(self, instance: _U, value: RepeatedNodeWrapper[_M]) -> None:
        repeated = self._inner_field.__get__(instance)
        replace_node(repeated, value.repeated)
        self._inner_field.__set__(instance, value.repeated)
        instance.__dict__[self._attr] = value


def _default_fset(instance: _U, value: _V) -> None:
    raise NotImplementedError()


class custom_property(base_rw_property[_V, _U]):
    def __init__(self, fget: Callable[[_U], _V]) -> None:
        super().__init__()
        self._fget = fget
        self._fset: Callable[[_U, _V], None] = _default_fset

    def setter(self, fset: Callable[[_U, _V], None]) -> 'custom_property[_V, _U]':
        self._fset = fset
        return self

    def _get(self, instance: _U) -> _V:
        return self._fget(instance)

    def __set__(self, instance: _U, value: _V) -> None:
        self._fset(instance, value)


class cached_custom_property(custom_property[_V, _U]):
    def __set_name__(self, owner: _U, name: str) -> None:
        self._attr = name

    def _get(self, instance: _U) -> _V:
        if self._attr in instance.__dict__:
            return instance.__dict__[self._attr]
        value = self._fget(instance)
        instance.__dict__[self._attr] = value
        return value

    def __set__(self, instance: _U, value: _V) -> None:
        super().__set__(instance, value)
        instance.__dict__[self._attr] = value


class unordered_node_property(base_rw_property[Optional[_V], _U]):
    def __init__(
            self,
            inner_property: base_rw_property[MutableSequence[_V | _M], _U],
            inner_type: Type[_V],
            *,
            prepend: bool = False,
    ) -> None:
        super().__init__()
        self._inner_property = inner_property
        self._inner_type = inner_type
        self._prepend = prepend

    def _get(self, instance: _U) -> Optional[_V]:
        wrapper = self._inner_property.__get__(instance)
        return next((item for item in wrapper if isinstance(item, self._inner_type)), None)

    def __set__(self, instance: _U, value: Optional[_V]) -> None:
        wrapper = self._inner_property.__get__(instance)
        i = next((i for i, item in enumerate(wrapper) if isinstance(item, self._inner_type)), None)
        if i is None and value is not None:
            if self._prepend:
                wrapper.insert(0, value)
            else:
                wrapper.append(value)
        elif i is not None and value is None:
            wrapper.pop(i)
        elif i is not None and value is not None:
            wrapper[i] = value



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import Optional, Type, TypeVar, final
from .. import base, internal, meta_value_internal
from ..block_comment import BlockComment
from ..inline_comment import InlineComment
from ..meta_key import MetaKey
from ..meta_value import MetaRawValue, MetaValue
from ..punctuation import Eol
from ..spacing import Newline, Whitespace

_Self = TypeVar('_Self', bound='Pushmeta')


@internal.token_model
class PushmetaLabel(internal.SimpleDefaultRawTokenModel):
    RULE = 'PUSHMETA'
    DEFAULT = 'pushmeta'


@internal.tree_model
class Pushmeta(internal.SurroundingCommentsMixin, base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'pushmeta'

    _label = internal.required_field[PushmetaLabel]()
    _key = internal.required_field[MetaKey]()
    _value = internal.optional_left_field[MetaRawValue](separators=(Whitespace.from_default(),))
    _inline_comment = internal.optional_left_field[InlineComment](separators=(Whitespace.from_default(),))
    _eol = internal.required_field[Eol]()

    @internal.custom_property
    def _leading_comment_pivot(self) -> base.RawTokenModel:
        return self._label.first_token

    @internal.custom_property
    def _value_pivot(self) -> base.RawTokenModel:
        return self._key.last_token

    @internal.custom_property
    def _inline_comment_pivot(self) -> base.RawTokenModel:
        return (self._value and self._value.last_token) or self._key.last_token

    @internal.custom_property
    def _trailing_comment_pivot(self) -> base.RawTokenModel:
        return self._eol.last_token

    raw_leading_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._leading_comment, _leading_comment_pivot)
    raw_key = internal.required_node_property(_key)
    raw_value = internal.optional_node_property(_value, _value_pivot)
    raw_inline_comment = internal.optional_node_property(_inline_comment, _inline_comment_pivot)
    raw_trailing_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._trailing_comment, _trailing_comment_pivot)

    leading_comment = internal.optional_string_property(raw_leading_comment, BlockComment)
    key = internal.required_value_property(raw_key)
    value = meta_value_internal.optional_meta_value_property(raw_value)
    inline_comment = internal.optional_string_property(raw_inline_comment, InlineComment)
    trailing_comment = internal.optional_string_property(raw_trailing_comment, BlockComment)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            leading_comment: Optional[BlockComment],
            label: PushmetaLabel,
            key: MetaKey,
            value: Optional[MetaRawValue],
            inline_comment: Optional[InlineComment],
            eol: Eol,
            trailing_comment: Optional[BlockComment],
    ):
        super().__init__(token_store)
        self._leading_comment = leading_comment
        self._label = label
        self._key = key
        self._value = value
        self._inline_comment = inline_comment
        self._eol = eol
        self._trailing_comment = trailing_comment

    @property
    def first_token(self) -> base.RawTokenModel:
        return (self._leading_comment and self._leading_comment.first_token) or self._label.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return (self._trailing_comment and self._trailing_comment.last_token) or self._eol.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._leading_comment.clone(self._leading_comment, token_store, token_transformer),
            type(self)._label.clone(self._label, token_store, token_transformer),
            type(self)._key.clone(self._key, token_store, token_transformer),
            type(self)._value.clone(self._value, token_store, token_transformer),
            type(self)._inline_comment.clone(self._inline_comment, token_store, token_transformer),
            type(self)._eol.clone(self._eol, token_store, token_transformer),
            type(self)._trailing_comment.clone(self._trailing_comment, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._leading_comment = type(self)._leading_comment.reattach(self._leading_comment, token_store, token_transformer)
        self._label = type(self)._label.reattach(self._label, token_store, token_transformer)
        self._key = type(self)._key.reattach(self._key, token_store, token_transformer)
        self._value = type(self)._value.reattach(self._value, token_store, token_transformer)
        self._inline_comment = type(self)._inline_comment.reattach(self._inline_comment, token_store, token_transformer)
        self._eol = type(self)._eol.reattach(self._eol, token_store, token_transformer)
        self._trailing_comment = type(self)._trailing_comment.reattach(self._trailing_comment, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, Pushmeta)
            and self._leading_comment == other._leading_comment
            and self._label == other._label
            and self._key == other._key
            and self._value == other._value
            and self._inline_comment == other._inline_comment
            and self._eol == other._eol
            and self._trailing_comment == other._trailing_comment
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            key: MetaKey,
            value: Optional[MetaRawValue],
            *,
            leading_comment: Optional[BlockComment] = None,
            inline_comment: Optional[InlineComment] = None,
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        label = PushmetaLabel.from_default()
        eol = Eol.from_default()
        tokens = [
            *cls._leading_comment.detach_with_separators(leading_comment),
            *label.detach(),
            Whitespace.from_default(),
            *key.detach(),
            *cls._value.detach_with_separators(value),
            *cls._inline_comment.detach_with_separators(inline_comment),
            *eol.detach(),
            *cls._trailing_comment.detach_with_separators(trailing_comment),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._leading_comment.reattach(leading_comment, token_store)
        cls._label.reattach(label, token_store)
        cls._key.reattach(key, token_store)
        cls._value.reattach(value, token_store)
        cls._inline_comment.reattach(inline_comment, token_store)
        cls._eol.reattach(eol, token_store)
        cls._trailing_comment.reattach(trailing_comment, token_store)
        return cls(token_store, leading_comment, label, key, value, inline_comment, eol, trailing_comment)

    @classmethod
    def from_value(
            cls: Type[_Self],
            key: str,
            value: Optional[MetaValue | MetaRawValue],
            *,
            leading_comment: Optional[str] = None,
            inline_comment: Optional[str] = None,
            trailing_comment: Optional[str] = None,
    ) -> _Self:
        return cls.from_children(
            leading_comment=BlockComment.from_value(leading_comment) if leading_comment is not None else None,
            key=MetaKey.from_value(key),
            value=meta_value_internal.from_value(value) if value is not None else None,
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            trailing_comment=BlockComment.from_value(trailing_comment) if trailing_comment is not None else None,
        )

    def auto_claim_comments(self) -> None:
        self.claim_leading_comment(ignore_if_already_claimed=True)
        self.claim_trailing_comment(ignore_if_already_claimed=True)
        type(self)._trailing_comment.auto_claim_comments(self._trailing_comment)
        type(self)._inline_comment.auto_claim_comments(self._inline_comment)
        type(self)._value.auto_claim_comments(self._value)
        type(self)._key.auto_claim_comments(self._key)
        type(self)._leading_comment.auto_claim_comments(self._leading_comment)



import datetime
import decimal
from typing import Union
from .account import Account
from .amount import Amount
from .bool import Bool
from .currency import Currency
from .date import Date
from .escaped_string import EscapedString
from .null import Null
from .number_expr import NumberExpr
from .tag import Tag

MetaRawValue = Account | Amount | Bool | Currency | Date | EscapedString | Null | NumberExpr | Tag
_ValueTypeSimplified = str | datetime.date | decimal.Decimal | bool
_ValueTypePreserved = Account | Currency | Tag | Null | Amount
MetaValue = Union[_ValueTypeSimplified, _ValueTypePreserved]



import re
from typing import Type, TypeVar, final
from . import base
from .internal import registry as _registry
from .internal import value_properties as _value_properties

_INDENT_SPLIT_RE = re.compile(r';\s*')
_Self = TypeVar('_Self', bound='BlockComment')


@_registry.token_model
class BlockComment(base.RawTokenModel, _value_properties.RWValueWithIndent[str]):
    RULE = 'BLOCK_COMMENT'

    @final
    def __init__(self, raw_text: str, indent: str, value: str, *, claimed: bool = True) -> None:
        super().__init__(raw_text)
        self._value = value
        self._indent = indent
        self._claimed = claimed

    @property
    def raw_text(self) -> str:
        return super().raw_text

    @raw_text.setter
    def raw_text(self, raw_text: str) -> None:
        self._update_raw_text(raw_text)
        self._indent, self._value = self._parse_value(raw_text)

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value
        self._update_raw_text(self._format_value(self._indent, value))

    @property
    def indent(self) -> str:
        return self._indent

    @indent.setter
    def indent(self, indent: str) -> None:
        self._indent = indent
        self._update_raw_text(self._format_value(indent, self._value))

    @property
    def claimed(self) -> bool:
        return self._claimed

    @claimed.setter
    def claimed(self, claimed: bool) -> None:
        self._claimed = claimed

    @classmethod
    def from_value(cls: Type[_Self], value: str, *, indent: str = '') -> _Self:
        return cls(cls._format_value(indent, value), indent, value)

    @classmethod
    def from_raw_text(cls: Type[_Self], raw_text: str) -> _Self:
        indent, value = cls._parse_value(raw_text)
        return cls(raw_text, indent, value)

    @classmethod
    def _parse_value(cls, raw_text: str) -> tuple[str, str]:
        indents, values = zip(*(
            tuple(_INDENT_SPLIT_RE.split(line, maxsplit=1))
            for line in raw_text.splitlines(keepends=True)
        ))
        return indents[0], ''.join(values)

    @classmethod
    def _format_value(cls, indent: str, value: str) -> str:
        return ''.join(
            f'{indent}; {line}' if line else f'{indent};'
            for line in value.splitlines(keepends=True) or ['']
        )

    def _clone(self: 'BlockComment') -> 'BlockComment':
        return type(self)(self.raw_text, self.indent, self.value, claimed=self.claimed)



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import Iterable, Type, TypeVar, final, get_args
from .. import base, internal
from ..balance import Balance
from ..block_comment import BlockComment
from ..close import Close
from ..commodity import Commodity
from ..custom import Custom
from ..document import Document
from ..event import Event
from ..ignored_line import IgnoredLine
from ..include import Include
from ..note import Note
from ..open import Open
from ..option import Option
from ..pad import Pad
from ..plugin import Plugin
from ..popmeta import Popmeta
from ..poptag import Poptag
from ..price import Price
from ..pushmeta import Pushmeta
from ..pushtag import Pushtag
from ..query import Query
from ..spacing import Newline
from ..transaction import Transaction

Directive = Balance | Close | Commodity | Custom | Document | Event | IgnoredLine | Include | Note | Open | Option | Pad | Plugin | Popmeta | Poptag | Price | Pushmeta | Pushtag | Query | Transaction
_Self = TypeVar('_Self', bound='File')


@internal.tree_model
class File(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'file'

    _directives = internal.repeated_field[Directive | BlockComment](separators=(Newline.from_default(), Newline.from_default()), separators_before=())

    raw_directives_with_comments = internal.repeated_node_with_interleaving_comments_property[Directive | BlockComment](_directives)
    raw_directives = internal.repeated_filtered_node_property[Directive](raw_directives_with_comments, get_args(Directive))

    directives = raw_directives

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            repeated_directives: internal.Repeated[Directive | BlockComment],
    ):
        super().__init__(token_store)
        self._directives = repeated_directives

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._directives.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._directives.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._directives.clone(self._directives, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._directives = type(self)._directives.reattach(self._directives, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, File)
            and self._directives == other._directives
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            directives: Iterable[Directive | BlockComment],
    ) -> _Self:
        repeated_directives = cls._directives.create_repeated(directives)
        tokens = [
            *cls._directives.detach_with_separators(repeated_directives),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._directives.reattach(repeated_directives, token_store)
        return cls(token_store, repeated_directives)

    @classmethod
    def from_value(
            cls: Type[_Self],
            directives: Iterable[Directive],
    ) -> _Self:
        return cls.from_children(
            directives=directives,
        )

    def auto_claim_comments(self) -> None:
        self.raw_directives_with_comments.auto_claim_comments()



import abc
import copy
from typing import Iterable, Optional, Type, TypeVar
from .. import base
from .repeated import Repeated
from .base_property import base_rw_property

_V = TypeVar('_V')
_M = TypeVar('_M', bound=base.RawModel)


class field(base_rw_property[_V, base.RawTreeModel]):

    def __set_name__(self, owner: Type[base.RawTreeModel], name: str) -> None:
        self._attr = name

    def _get(self, instance: base.RawTreeModel) -> _V:
        return instance.__dict__[self._attr]

    def __set__(self, instance: base.RawTreeModel, value: _V) -> None:
        instance.__dict__[self._attr] = value

    @abc.abstractmethod
    def clone(
            self,
            value: _V,
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer,
    ) -> _V:
        ...

    @abc.abstractmethod
    def reattach(
            self,
            value: _V,
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer = base.IDENTITY_TOKEN_TRANSFORMER,
    ) -> _V:
        ...

    @abc.abstractmethod
    def auto_claim_comments(self, value: _V) -> None:
        ...

    @abc.abstractmethod
    def detach_with_separators(self, value: _V) -> list[base.RawTokenModel]:
        ...


class required_field(field[_M]):

    def clone(
            self,
            value: _M,
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer,
    ) -> _M:
        return value.clone(token_store, token_transformer)

    def reattach(
            self,
            value: _M,
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer = base.IDENTITY_TOKEN_TRANSFORMER,
    ) -> _M:
        return value.reattach(token_store, token_transformer)

    def auto_claim_comments(self, value: _M) -> None:
        value.auto_claim_comments()

    def detach_with_separators(self, value: _M) -> list[base.RawTokenModel]:
        return value.detach()


class optional_field(field[Optional[_M]]):

    def __init__(self, *, separators: tuple[base.RawTokenModel, ...]) -> None:
        super().__init__()
        self._separators = separators

    @property
    def separators(self) -> tuple[base.RawTokenModel, ...]:
        return self._separators

    def clone(
            self,
            value: Optional[_M],
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer,
    ) -> Optional[_M]:
        return value.clone(token_store, token_transformer) if value is not None else None

    def reattach(
            self,
            value: Optional[_M],
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer = base.IDENTITY_TOKEN_TRANSFORMER,
    ) -> Optional[_M]:
        return value.reattach(token_store, token_transformer) if value is not None else None

    def auto_claim_comments(self, value: Optional[_M]) -> None:
        if value is not None:
            value.auto_claim_comments()

    @abc.abstractmethod
    def _create_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            value: _M,
    ) -> None:
        ...

    @abc.abstractmethod
    def _remove_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            current: _M,
    ) -> None:
        ...


class optional_left_field(optional_field[_M]):

    def detach_with_separators(self, value: Optional[_M]) -> list[base.RawTokenModel]:
        return [*copy.deepcopy(self.separators), *value.detach()] if value is not None else []

    def _create_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            value: _M,
    ) -> None:
        token_store.insert_after(pivot, [
            *copy.deepcopy(self.separators),
            *value.detach(),
        ])
        value.reattach(token_store)

    def _remove_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            current: _M,
    ) -> None:
        first = token_store.get_next(pivot)
        assert first is not None
        token_store.remove(first, current.last_token)


class optional_right_field(optional_field[_M]):

    def detach_with_separators(self, value: Optional[_M]) -> list[base.RawTokenModel]:
        return [*value.detach(), *copy.deepcopy(self.separators)] if value is not None else []

    def _create_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            value: _M,
    ) -> None:
        token_store.insert_before(pivot, [
            *value.detach(),
            *copy.deepcopy(self.separators),
        ])
        value.reattach(token_store)

    def _remove_node(
            self,
            token_store: base.TokenStore,
            pivot: base.RawTokenModel,
            current: _M,
    ) -> None:
        last = token_store.get_prev(pivot)
        assert last is not None
        token_store.remove(current.first_token, last)


class repeated_field(field[Repeated[_M]]):
    def __init__(
            self,
            *,
            separators: tuple[base.RawTokenModel, ...],
            separators_before: Optional[tuple[base.RawTokenModel, ...]] = None,
            default_indent: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._separators = separators
        self._separators_before = separators_before
        self._default_indent = default_indent

    @property
    def separators(self) -> tuple[base.RawTokenModel, ...]:
        return self._separators

    @property
    def separators_before(self) -> Optional[tuple[base.RawTokenModel, ...]]:
        return self._separators_before

    @property
    def default_indent(self) -> Optional[str]:
        return self._default_indent

    def create_repeated(self, values: Iterable[_M]) -> Repeated[_M]:
        return Repeated.from_children(
            values,
            separators=self.separators,
            separators_before=self.separators_before,
            indent=self.default_indent)

    def clone(
            self,
            value: Repeated[_M],
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer,
    ) -> Repeated[_M]:
        return value.clone(token_store, token_transformer)

    def reattach(
            self,
            value: Repeated[_M],
            token_store: base.TokenStore,
            token_transformer: base.TokenTransformer = base.IDENTITY_TOKEN_TRANSFORMER,
    ) -> Repeated[_M]:
        return value.reattach(token_store, token_transformer)

    def auto_claim_comments(self, value: Repeated[_M]) -> None:
        value.auto_claim_comments()

    def detach_with_separators(self, value: Repeated[_M]) -> list[base.RawTokenModel]:
        return value.detach()



import io
from typing import TypeVar
from autobean_refactor import models

_T = TypeVar('_T', bound=io.TextIOBase)


def print_model(model: models.RawModel, file: _T) -> _T:
    for token in model.tokens:
        file.write(token.raw_text)
    return file



import datetime
import itertools
from typing import Iterable, Mapping, Optional, Type, TypeVar
from . import base, internal, meta_item_internal
from .block_comment import BlockComment
from .date import Date
from .escaped_string import EscapedString
from .inline_comment import InlineComment
from .link import Link
from .meta_value import MetaRawValue, MetaValue
from .meta_item import MetaItem
from .posting import Posting
from .tag import Tag
from .transaction_flag import TransactionFlag
from .generated import transaction

_Self = TypeVar('_Self', bound='Transaction')


@internal.tree_model
class Transaction(transaction.Transaction):

    @classmethod
    def from_parsed_children(cls: Type[_Self], token_store: base.TokenStore, *children: Optional[base.RawModel]) -> _Self:
        (
            leading_comment,
            date,
            flag,
            string0,
            string1,
            string2,
            *args,
        ) = children
        if string1 is not None and string2 is None:
            string1, string2 = string0, string1
        return super().from_parsed_children(
            token_store,
            leading_comment,
            date,
            flag,
            string0,
            string1,
            string2,
            *args)

    @internal.custom_property
    def raw_payee(self) -> Optional[EscapedString]:
        return self.raw_string1

    @raw_payee.setter
    def __raw_payee(self, value: Optional[EscapedString]) -> None:
        if value is not None and self.raw_narration is None:
            self.raw_narration = EscapedString.from_value('') 
        self.raw_string1 = value

    @internal.custom_property
    def raw_narration(self) -> Optional[EscapedString]:
        return self.raw_string2

    @raw_narration.setter
    def __raw_narration(self, value: Optional[EscapedString]) -> None:
        if value is None and self.raw_payee is not None:
            value = EscapedString.from_value('')
        self.raw_string2 = value

    payee = internal.optional_string_property(raw_payee, EscapedString)
    narration = internal.optional_string_property(raw_narration, EscapedString)
    tags = internal.repeated_string_property(transaction.Transaction.raw_tags_links, Tag)
    links = internal.repeated_string_property(transaction.Transaction.raw_tags_links, Link)

    @classmethod
    def from_children(  # type: ignore[override]
            cls: Type[_Self],
            date: Date,
            flag: TransactionFlag,
            payee: Optional[EscapedString],
            narration: Optional[EscapedString],
            postings: Iterable[Posting],
            *,
            leading_comment: Optional[BlockComment] = None,
            tags_links: Iterable[Link | Tag] = (),
            inline_comment: Optional[InlineComment] = None,
            meta: Iterable[MetaItem | BlockComment] = (),
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        if payee is not None and narration is None:
            narration = EscapedString.from_value('')
        return super().from_children(
            date,
            flag,
            None,
            payee,
            narration,
            postings,
            leading_comment=leading_comment,
            tags_links=tags_links,
            inline_comment=inline_comment,
            meta=meta,
            trailing_comment=trailing_comment,
        )

    @classmethod
    def from_value(
            cls: Type[_Self],
            date: datetime.date,
            payee: Optional[str],
            narration: Optional[str],
            postings: Iterable[Posting],
            *,
            flag: str = '*',
            tags: Iterable[str] = (),
            links: Iterable[str] = (),
            inline_comment: Optional[str] = None,
            meta: Optional[Mapping[str, MetaRawValue | MetaValue]] = None,
    ) -> _Self:
        return cls.from_children(
            date=Date.from_value(date),
            flag=TransactionFlag.from_value(flag),
            payee=EscapedString.from_value(payee) if payee is not None else None,
            narration=EscapedString.from_value(narration) if narration is not None else None,
            tags_links=itertools.chain(map(Tag.from_value, tags), map(Link.from_value, links)),
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            meta=meta_item_internal.from_mapping(meta) if meta is not None else (),
            postings=postings,
        )



from .internal import TOKEN_MODELS, TREE_MODELS
from .base import *
from .account import *
from .amount import *
from .balance import *
from .block_comment import *
from .bool import *
from .close import *
from .commodity import *
from .compound_amount import *
from .cost import *
from .cost_component import *
from .cost_spec import *
from .currency import *
from .custom import *
from .date import *
from .document import *
from .escaped_string import *
from .event import *
from .file import *
from .ignored import *
from .ignored_line import *
from .include import *
from .inline_comment import *
from .link import *
from .meta_item import *
from .meta_key import *
from .meta_value import *
from .note import *
from .null import *
from .number import *
from .number_add_expr import *
from .number_atom_expr import *
from .number_expr import *
from .number_mul_expr import *
from .number_paren_expr import *
from .number_unary_expr import *
from .open import *
from .option import *
from .pad import *
from .plugin import *
from .popmeta import *
from .poptag import *
from .posting import *
from .posting_flag import *
from .price import *
from .punctuation import *
from .pushmeta import *
from .pushtag import *
from .query import *
from .tag import *
from .tolerance import *
from .total_price import *
from .transaction import *
from .transaction_flag import *
from .unit_price import *



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import decimal
from typing import Iterable, Mapping, Optional, Type, TypeVar, final
from .. import base, internal, meta_item_internal
from ..account import Account
from ..block_comment import BlockComment
from ..cost_spec import CostSpec
from ..currency import Currency
from ..inline_comment import InlineComment
from ..meta_item import MetaItem
from ..meta_value import MetaRawValue, MetaValue
from ..number_expr import NumberExpr
from ..posting_flag import PostingFlag
from ..punctuation import Eol, Indent
from ..spacing import Newline, Whitespace
from ..total_price import TotalPrice
from ..unit_price import UnitPrice

PriceAnnotation = TotalPrice | UnitPrice
_Self = TypeVar('_Self', bound='Posting')


@internal.tree_model
class Posting(internal.SurroundingCommentsMixin, base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'posting'

    _indent = internal.required_field[Indent]()
    _flag = internal.optional_right_field[PostingFlag](separators=(Whitespace.from_default(),))
    _account = internal.required_field[Account]()
    _number = internal.optional_left_field[NumberExpr](separators=(Whitespace.from_default(),))
    _currency = internal.optional_left_field[Currency](separators=(Whitespace.from_default(),))
    _cost = internal.optional_left_field[CostSpec](separators=(Whitespace.from_default(),))
    _price = internal.optional_left_field[PriceAnnotation](separators=(Whitespace.from_default(),))
    _inline_comment = internal.optional_left_field[InlineComment](separators=(Whitespace.from_default(),))
    _eol = internal.required_field[Eol]()
    _meta = internal.repeated_field[MetaItem | BlockComment](separators=(Newline.from_default(),), default_indent='        ')

    @internal.custom_property
    def _leading_comment_pivot(self) -> base.RawTokenModel:
        return self._indent.first_token

    @internal.custom_property
    def _flag_pivot(self) -> base.RawTokenModel:
        return self._account.first_token

    @internal.custom_property
    def _number_pivot(self) -> base.RawTokenModel:
        return self._account.last_token

    @internal.custom_property
    def _currency_pivot(self) -> base.RawTokenModel:
        return (self._number and self._number.last_token) or self._account.last_token

    @internal.custom_property
    def _cost_pivot(self) -> base.RawTokenModel:
        return (self._currency and self._currency.last_token) or (self._number and self._number.last_token) or self._account.last_token

    @internal.custom_property
    def _price_pivot(self) -> base.RawTokenModel:
        return (self._cost and self._cost.last_token) or (self._currency and self._currency.last_token) or (self._number and self._number.last_token) or self._account.last_token

    @internal.custom_property
    def _inline_comment_pivot(self) -> base.RawTokenModel:
        return (self._price and self._price.last_token) or (self._cost and self._cost.last_token) or (self._currency and self._currency.last_token) or (self._number and self._number.last_token) or self._account.last_token

    @internal.custom_property
    def _trailing_comment_pivot(self) -> base.RawTokenModel:
        return self._meta.last_token or self._eol.last_token

    raw_leading_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._leading_comment, _leading_comment_pivot)
    raw_indent = internal.required_node_property(_indent)
    raw_flag = internal.optional_node_property(_flag, _flag_pivot)
    raw_account = internal.required_node_property(_account)
    raw_number = internal.optional_node_property(_number, _number_pivot)
    raw_currency = internal.optional_node_property(_currency, _currency_pivot)
    raw_cost = internal.optional_node_property(_cost, _cost_pivot)
    raw_price = internal.optional_node_property(_price, _price_pivot)
    raw_inline_comment = internal.optional_node_property(_inline_comment, _inline_comment_pivot)
    raw_meta_with_comments = internal.repeated_node_with_interleaving_comments_property(_meta)
    raw_meta = meta_item_internal.repeated_raw_meta_item_property(raw_meta_with_comments)
    raw_trailing_comment = internal.optional_node_property(internal.SurroundingCommentsMixin._trailing_comment, _trailing_comment_pivot)

    leading_comment = internal.optional_indented_string_property(raw_leading_comment, BlockComment, raw_indent)
    indent = internal.required_value_property(raw_indent)
    flag = internal.optional_string_property(raw_flag, PostingFlag)
    account = internal.required_value_property(raw_account)
    number = internal.optional_decimal_property(raw_number, NumberExpr)
    currency = internal.optional_string_property(raw_currency, Currency)
    cost = raw_cost
    price = raw_price
    inline_comment = internal.optional_string_property(raw_inline_comment, InlineComment)
    meta = meta_item_internal.repeated_meta_item_property(raw_meta_with_comments)
    trailing_comment = internal.optional_indented_string_property(raw_trailing_comment, BlockComment, raw_indent)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            leading_comment: Optional[BlockComment],
            indent: Indent,
            flag: Optional[PostingFlag],
            account: Account,
            number: Optional[NumberExpr],
            currency: Optional[Currency],
            cost: Optional[CostSpec],
            price: Optional[PriceAnnotation],
            inline_comment: Optional[InlineComment],
            eol: Eol,
            repeated_meta: internal.Repeated[MetaItem | BlockComment],
            trailing_comment: Optional[BlockComment],
    ):
        super().__init__(token_store)
        self._leading_comment = leading_comment
        self._indent = indent
        self._flag = flag
        self._account = account
        self._number = number
        self._currency = currency
        self._cost = cost
        self._price = price
        self._inline_comment = inline_comment
        self._eol = eol
        self._meta = repeated_meta
        self._trailing_comment = trailing_comment

    @property
    def first_token(self) -> base.RawTokenModel:
        return (self._leading_comment and self._leading_comment.first_token) or self._indent.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return (self._trailing_comment and self._trailing_comment.last_token) or self._meta.last_token or self._eol.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._leading_comment.clone(self._leading_comment, token_store, token_transformer),
            type(self)._indent.clone(self._indent, token_store, token_transformer),
            type(self)._flag.clone(self._flag, token_store, token_transformer),
            type(self)._account.clone(self._account, token_store, token_transformer),
            type(self)._number.clone(self._number, token_store, token_transformer),
            type(self)._currency.clone(self._currency, token_store, token_transformer),
            type(self)._cost.clone(self._cost, token_store, token_transformer),
            type(self)._price.clone(self._price, token_store, token_transformer),
            type(self)._inline_comment.clone(self._inline_comment, token_store, token_transformer),
            type(self)._eol.clone(self._eol, token_store, token_transformer),
            type(self)._meta.clone(self._meta, token_store, token_transformer),
            type(self)._trailing_comment.clone(self._trailing_comment, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._leading_comment = type(self)._leading_comment.reattach(self._leading_comment, token_store, token_transformer)
        self._indent = type(self)._indent.reattach(self._indent, token_store, token_transformer)
        self._flag = type(self)._flag.reattach(self._flag, token_store, token_transformer)
        self._account = type(self)._account.reattach(self._account, token_store, token_transformer)
        self._number = type(self)._number.reattach(self._number, token_store, token_transformer)
        self._currency = type(self)._currency.reattach(self._currency, token_store, token_transformer)
        self._cost = type(self)._cost.reattach(self._cost, token_store, token_transformer)
        self._price = type(self)._price.reattach(self._price, token_store, token_transformer)
        self._inline_comment = type(self)._inline_comment.reattach(self._inline_comment, token_store, token_transformer)
        self._eol = type(self)._eol.reattach(self._eol, token_store, token_transformer)
        self._meta = type(self)._meta.reattach(self._meta, token_store, token_transformer)
        self._trailing_comment = type(self)._trailing_comment.reattach(self._trailing_comment, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, Posting)
            and self._leading_comment == other._leading_comment
            and self._indent == other._indent
            and self._flag == other._flag
            and self._account == other._account
            and self._number == other._number
            and self._currency == other._currency
            and self._cost == other._cost
            and self._price == other._price
            and self._inline_comment == other._inline_comment
            and self._eol == other._eol
            and self._meta == other._meta
            and self._trailing_comment == other._trailing_comment
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            account: Account,
            number: Optional[NumberExpr],
            currency: Optional[Currency],
            *,
            leading_comment: Optional[BlockComment] = None,
            indent: Indent,
            flag: Optional[PostingFlag] = None,
            cost: Optional[CostSpec] = None,
            price: Optional[PriceAnnotation] = None,
            inline_comment: Optional[InlineComment] = None,
            meta: Iterable[MetaItem | BlockComment] = (),
            trailing_comment: Optional[BlockComment] = None,
    ) -> _Self:
        eol = Eol.from_default()
        repeated_meta = cls._meta.create_repeated(meta)
        tokens = [
            *cls._leading_comment.detach_with_separators(leading_comment),
            *indent.detach(),
            *cls._flag.detach_with_separators(flag),
            *account.detach(),
            *cls._number.detach_with_separators(number),
            *cls._currency.detach_with_separators(currency),
            *cls._cost.detach_with_separators(cost),
            *cls._price.detach_with_separators(price),
            *cls._inline_comment.detach_with_separators(inline_comment),
            *eol.detach(),
            *cls._meta.detach_with_separators(repeated_meta),
            *cls._trailing_comment.detach_with_separators(trailing_comment),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._leading_comment.reattach(leading_comment, token_store)
        cls._indent.reattach(indent, token_store)
        cls._flag.reattach(flag, token_store)
        cls._account.reattach(account, token_store)
        cls._number.reattach(number, token_store)
        cls._currency.reattach(currency, token_store)
        cls._cost.reattach(cost, token_store)
        cls._price.reattach(price, token_store)
        cls._inline_comment.reattach(inline_comment, token_store)
        cls._eol.reattach(eol, token_store)
        cls._meta.reattach(repeated_meta, token_store)
        cls._trailing_comment.reattach(trailing_comment, token_store)
        return cls(token_store, leading_comment, indent, flag, account, number, currency, cost, price, inline_comment, eol, repeated_meta, trailing_comment)

    @classmethod
    def from_value(
            cls: Type[_Self],
            account: str,
            number: Optional[decimal.Decimal],
            currency: Optional[str],
            *,
            leading_comment: Optional[str] = None,
            indent: str = '    ',
            flag: Optional[str] = None,
            cost: Optional[CostSpec] = None,
            price: Optional[PriceAnnotation] = None,
            inline_comment: Optional[str] = None,
            meta: Optional[Mapping[str, MetaValue | MetaRawValue]] = None,
            trailing_comment: Optional[str] = None,
    ) -> _Self:
        return cls.from_children(
            leading_comment=BlockComment.from_value(leading_comment) if leading_comment is not None else None,
            indent=Indent.from_value(indent),
            flag=PostingFlag.from_value(flag) if flag is not None else None,
            account=Account.from_value(account),
            number=NumberExpr.from_value(number) if number is not None else None,
            currency=Currency.from_value(currency) if currency is not None else None,
            cost=cost,
            price=price,
            inline_comment=InlineComment.from_value(inline_comment) if inline_comment is not None else None,
            meta=meta_item_internal.from_mapping(meta) if meta is not None else (),
            trailing_comment=BlockComment.from_value(trailing_comment) if trailing_comment is not None else None,
        )

    def auto_claim_comments(self) -> None:
        self.claim_leading_comment(ignore_if_already_claimed=True)
        self.claim_trailing_comment(ignore_if_already_claimed=True)
        type(self)._trailing_comment.auto_claim_comments(self._trailing_comment)
        self.raw_meta_with_comments.auto_claim_comments()
        type(self)._inline_comment.auto_claim_comments(self._inline_comment)
        type(self)._price.auto_claim_comments(self._price)
        type(self)._cost.auto_claim_comments(self._cost)
        type(self)._currency.auto_claim_comments(self._currency)
        type(self)._number.auto_claim_comments(self._number)
        type(self)._account.auto_claim_comments(self._account)
        type(self)._flag.auto_claim_comments(self._flag)
        type(self)._indent.auto_claim_comments(self._indent)
        type(self)._leading_comment.auto_claim_comments(self._leading_comment)



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

from typing import TYPE_CHECKING, Type, TypeVar, final
from .. import base, internal
if TYPE_CHECKING:
  from ..number_add_expr import NumberAddExpr

_Self = TypeVar('_Self', bound='NumberExpr')


@internal.tree_model
class NumberExpr(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'number_expr'

    _number_add_expr = internal.required_field['NumberAddExpr']()

    raw_number_add_expr = internal.required_node_property(_number_add_expr)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            number_add_expr: 'NumberAddExpr',
    ):
        super().__init__(token_store)
        self._number_add_expr = number_add_expr

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._number_add_expr.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._number_add_expr.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._number_add_expr.clone(self._number_add_expr, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._number_add_expr = type(self)._number_add_expr.reattach(self._number_add_expr, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, NumberExpr)
            and self._number_add_expr == other._number_add_expr
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            number_add_expr: 'NumberAddExpr',
    ) -> _Self:
        tokens = [
            *number_add_expr.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._number_add_expr.reattach(number_add_expr, token_store)
        return cls(token_store, number_add_expr)

    def auto_claim_comments(self) -> None:
        type(self)._number_add_expr.auto_claim_comments(self._number_add_expr)



from . import internal


@internal.token_model
class MetaKey(internal.SingleValueRawTokenModel[str]):
    RULE = 'META_KEY'

    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        return raw_text[:-1]

    @classmethod
    def _format_value(cls, value: str) -> str:
        return f'{value}:'
